/*
The screen is our viewing window on the world. This class has several
variables:

There are two final class variables of type Screen, one for the main window and
one for the animation window.

Then there a several properties of the screen:
- The screen has a point in the middle which is the vector s^m. We implement
this as the Vector object screenMiddle.
- At some distance from the screen there is the retina, denoted by the vector r
and this is the Vector object retina.
- r is placed from a distance delta to the screen. We represent delta by float
RETINA_SCREEN_DISTANCE.
- If u, v and n are a negatively oriented orthonormal base where u and v span
the screen, then going from r in the direction of n, we travel a distance of
delta and end up in s^m. In other words, s^m - r is orthogonal to the screen
and has a length of delta: s^m - r = delta n.
u, v and n are the Vector objects u, v and normal respectively.
- The screen has a width in pixels, this is int SCREEN_WIDTH.
- The screen has a height in pixels, this is int SCREEN_HEIGHT.

The properties s^m, r, u, v and n are defined twice: as final class variables
and as non-final instance variables. The non-final variables are normally used
but when a new tree file is read, then for the main window the viewpoint is
reset and the non-final variables are reset to their final equivalents.

Note that for the main window the initial values for s^m and r differ from
their initial viewpoint equivalents. This is because at start up the program
plays a small animation in which the screen is in a different start position
(u, v and n are the same though by "coincidence").
*/

package display;

import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import java.util.HashSet;
import geometry.Vector;
import tree.Branch;
import tree.Branches;
import tree.Surface;
import tree.Surfaces;
import visual_aid.Marker;
import visual_aid.Markers;
import visual_aid.Trail;
import visual_aid.TrailSegment;


public class Screen {
	
	// scalars and vectors:
	
	public static final Vector SCREEN_MIDDLE = new Vector(0.0f, 0.0f, 0.0f);
	public Vector screenMiddle;
	// s^m
	
	public static final float RETINA_SCREEN_DISTANCE = 1000.0f;
	// delta
	
	public static final Vector NORMAL = new Vector(0.0f, 0.0f, 1.0f);
	public Vector normal;
	// n
	
	public static final Vector RETINA =
	 SCREEN_MIDDLE.subtract(NORMAL.multiply(RETINA_SCREEN_DISTANCE));
	public Vector retina;
	// r
	
	public static final Vector U = new Vector(1.0f, 0.0f, 0.0f);
	public Vector u;
	// u
	
	public static final Vector V = new Vector(0.0f, 1.0f, 0.0f);
	public Vector v;
	// v
	
	public static final int SCREEN_WIDTH = 960;
	public static final int SCREEN_HEIGHT = 720;
	
	// main window and animation screens
	public static final Screen MAIN_SCREEN =
	 new Screen(new Vector(0.0f, 2000.0f, -5000.0f), Screen.NORMAL, Screen.U,
	 Screen.V), ANIMATION_SCREEN =
	 new Screen(Screen.SCREEN_MIDDLE, Screen.NORMAL, Screen.U, Screen.V);
	 // for ANIMATION_SCREEN we just want the variable created,
	 // any value would do
	
	private HashSet<OpaquePolygon> opaquePolygons; // the screen representation
												   // of all surfaces
												   // for this screen
	
	private Lock oPLock = new ReentrantLock(); // lock on behalf of the above
											   // hash set of opaque polygons
	
	public Screen(Vector screenMiddle, Vector normal, Vector u, Vector v) {
		this.screenMiddle= screenMiddle;
		this.normal = normal;
		retina = this.screenMiddle.subtract(this.normal.multiply(RETINA_SCREEN_DISTANCE));
		this.u = u;
		this.v = v;
	}
	
	public HashSet<OpaquePolygon> getOpaquePolygons() {
		return opaquePolygons;
	}
	
	public Lock getOPLock() {
		return oPLock;
	}
	
	// method to reset the viewpoint to the initial viewpoint
	public void resetViewpoint() {
		screenMiddle = SCREEN_MIDDLE;
		normal = NORMAL;
		retina = RETINA;
		u = U;
		v = V;
	}
	
	// method to determine the visibility of a given vector
	// for the screen plane method
	// (the retina plane is on the non-visible side of the screen plane)
	public Visibility visibilitySPM(Vector p) {
		float innerProduct =
		 p.subtract(screenMiddle).innerProduct(normal);
		if (innerProduct > 0)
			return Visibility.VISIBLE;
		else if (innerProduct < 0)
			return Visibility.INVISIBLE;
		else
			return Visibility.ON_PLANE;
	}
	
	// method to determine the visibility of a given vector
	// for the retina plane method
	// (the screen plane is on the visible side of the retina plane)
	public Visibility visibilityRPM(Vector p) {
		float innerProduct =
		 p.subtract(retina).innerProduct(normal);
		if (innerProduct > 0)
			return Visibility.VISIBLE;
		else if (innerProduct < 0)
			return Visibility.INVISIBLE;
		else
			return Visibility.ON_PLANE;
	}
	
	
	// method for finding the screen representation of every single branch,
	// marker, trail segment and surface
	public void findRepresentation() {
		Branches branches = Branches.getInstance();
		for (Branch branch : branches.values()) {
			branch.findScreenRepresentation(this);
		}
		Markers markers = Markers.getInstance();
		for (Marker marker : markers) {
			marker.findScreenRepresentation(this);
		}
		Trail trail = Trail.getInstance();
		for (TrailSegment trailSegment : trail) {
			trailSegment.findScreenRepresentation(this);
		}
		
		if (oPLock.tryLock()) { // if not locked (by other thread), set the lock
			try {
				Surfaces surfaces = Surfaces.getInstance();
				opaquePolygons = new HashSet<OpaquePolygon>();
				for (Surface surface : surfaces) {
					OpaquePolygon opaquePolygon = surface.findScreenRepresentation(this);
					if (opaquePolygon != null) {
						opaquePolygons.add(opaquePolygon);
					}
				}
			} finally {
				oPLock.unlock(); // whatever happens, always unlock (standard mechanism)
			}
		}
	}
	
}
