/*
The PlanarPolygon class represents a polygon in the two dimensional space.
*/

package geometry;

import java.util.ArrayList;
import display.Screen;

public class PlanarPolygon extends Polygon<PlanarVector> {
	
	public PlanarPolygon(ArrayList<PlanarVector> vertices) {
		super(vertices);
	}
	
	// method to clip the planar polygon (the subject polygon) to the screen
	// the method follows the Sutherland-Hodgman algorithm
	// for details: see the Functional Description of tree package
	// pre-condition: the planar polygon has been decluttered
	public PlanarPolygon clipToScreen() {
		
		PlanarPolygon subjectPolygon;
		PlanarVector a, b; // the subject polygon is clipped to the four sides
						   // of the screen: the bottom side, the right side,
						   // the top side and the left side
						   // the vectors a and b will be the starting point
						   // and the direction of these sides
		
		subjectPolygon = this; // start with this planar polygon
		
		a = new PlanarVector(-Screen.SCREEN_WIDTH / 2.0f,
							 -Screen.SCREEN_HEIGHT / 2.0f);
		b = new PlanarVector(1.0f, 0.0f);
		subjectPolygon = subjectPolygon.clipToSide(a, b); // clip to one side,
														  // notice that
														  // the result will be
														  // the new subject
		subjectPolygon.declutter(); // always declutter
		
		// do the same for the other sides
		a = new PlanarVector(Screen.SCREEN_WIDTH / 2.0f,
							 -Screen.SCREEN_HEIGHT / 2.0f);
		b = new PlanarVector(0.0f, 1.0f);
		subjectPolygon = subjectPolygon.clipToSide(a, b);
		subjectPolygon.declutter();
		
		a = new PlanarVector(Screen.SCREEN_WIDTH / 2.0f,
							 Screen.SCREEN_HEIGHT / 2.0f);
		b = new PlanarVector(-1.0f, 0.0f);
		subjectPolygon = subjectPolygon.clipToSide(a, b);
		subjectPolygon.declutter();
		
		a = new PlanarVector(-Screen.SCREEN_WIDTH / 2.0f,
							 Screen.SCREEN_HEIGHT / 2.0f);
		b = new PlanarVector(0.0f, -1.0f);
		subjectPolygon = subjectPolygon.clipToSide(a, b);
		subjectPolygon.declutter();
		
		return subjectPolygon;
	}
	
	// method to clip the planar polygon (the subject polygon) to one side of
	// the screen
	// the method follows the Sutherland-Hodgman algorithm
	// for details: see the Functional Description of tree package
	private PlanarPolygon clipToSide(PlanarVector ap, PlanarVector bp) {
	 // a + lambda b is used for the side of the screen
	 // note: p means planar
		
		ArrayList<PlanarVector> newVertices = new ArrayList<PlanarVector>();
		 // the vertices of the new subject polygon
		
		if (vertices.size() >= 2) { // only consider polygons with at least 2
									// vertices
			
			PlanarVector previousVertix = vertices.get(vertices.size() - 1);
			Vector d = previousVertix.subtract(ap); // d is node_last - a
			Vector normal = bp.crossProduct(d);
			boolean isPreviousVertixInside = (normal.getZ() >= 0.0f);
			
			for (PlanarVector cp : vertices) {
				d = cp.subtract(ap); // d is node_i - a
				normal = bp.crossProduct(d);
				
				if ((normal.getZ() >= 0.0f) & isPreviousVertixInside) {
				 // case: this vertix is inside and so was the previous
				 // so add the current vertix to the list
					newVertices.add(cp);
				} else if (!(normal.getZ() >= 0.0f) & !isPreviousVertixInside) {
				 // case: this vertix is outside and so was the previous
				 // so do not add anything to the list
				} else {
				 // case: this vertix is inside and the previous was outside or
				 // vice versa so we have to add the intersection to the list,
				 // we have to solve a + lambda b = node_i-1 + mu (node_i - node_i-1)
					Vector e = cp.subtract(previousVertix);
					PlanarVector ep = new PlanarVector(e.getX(), e.getY());
					PlanarVector lambdaMu =
					 PlanarVector.calculateIntersection(ap, bp, previousVertix, ep);
					Vector f = ap.add(bp.multiply(lambdaMu.getX()));
					PlanarVector fp = new PlanarVector(f.getX(), f.getY());
					newVertices.add(fp); // now add the intersection to the list
					isPreviousVertixInside = !isPreviousVertixInside;
					if (normal.getZ() >= 0.0f) {
					 // we may have to add the current vertix to the list as well
						newVertices.add(cp);
					}
				}
				previousVertix = cp;
			}
			
		}
		
		return new PlanarPolygon(newVertices); // create a planar polygon from the list
	}
	
}
