/*
The Rotation class represents a rotation in the three dimensional space.
Every rotation can be represented by
- an axis of rotation, the vector n, and
- a rotation angle, the scalar theta,
where n is a unit vector, i.e. a vector of length 1, and 0 <= theta <= pi.

The convertToMatrix() and convertFromMatrix() methods make extensive use of
Rodrigues' formula. See:
https://www2.cs.duke.edu/courses/fall13/compsci527/notes/rodrigues.pdf
*/

package geometry;

public class Rotation {
	
	public static final float EPSILON = 0.001f; // rounding margin
	
	private final Vector n;    // once a Rotation object has been created, it
	private final float theta; // cannot be changed
	
	public Rotation(Vector n, float theta) {
		this.n = n;
		this.theta = theta;
	}
	
	public Vector getN() {
		return n;
	}
	
	public float getTheta() {
		return theta;
	}
	
	// method to calculate the matrix representation of this rotation using
	// Rodrigues' formula
	public Matrix convertToMatrix() {
		Vector c1, c2, c3;
		
		float n_x = n.getX();
		float n_y = n.getY();
		float n_z = n.getZ();
		float cos0 = (float)Math.cos(theta); // the '0' is for theta
		float sin0 = (float)Math.sin(theta);
		
		c1 = new Vector(cos0 + n_x * n_x * (1.0f - cos0),
						n_x * n_y * (1.0f - cos0) + n_z * sin0,
						n_x * n_z * (1.0f - cos0) - n_y * sin0);
		
		c2 = new Vector(n_x * n_y * (1.0f - cos0) - n_z * sin0,
						cos0 + n_y * n_y * (1.0f - cos0),
						n_y * n_z * (1.0f - cos0) + n_x * sin0);
		
		c3 = new Vector(n_x * n_z * (1.0f - cos0) + n_y * sin0,
						n_y * n_z * (1.0f - cos0) - n_x * sin0,
						cos0 + n_z * n_z * (1.0f - cos0));
		
		Matrix r = new Matrix(c1, c2, c3);
		return r;
	}
	
	// method to calculate the axis and angle of a rotation from its matrix
	// representation using Rodrigues' formula
	public static Rotation convertFromMatrix(Matrix r) { // it is the caller's
														 // responsibility to
														 // provide a matrix
														 // which is a rotation
		float trace = r.trace();
		float cos0 = (trace - 1.0f) / 2.0f;
		
		Vector n;
		float theta;
		
		if (cos0 >= -1.0f + EPSILON & cos0 <= 1.0f - EPSILON) {
			// theta is well between 0 and pi
			theta = (float)Math.acos(cos0);
			n = new Vector(r.getC2().getZ() - r.getC3().getY(),
						   r.getC3().getX() - r.getC1().getZ(),
						   r.getC1().getY() - r.getC2().getX());
			// we have to normalize n
			n = n.multiply(1.0f/n.length());
			
		} else if (cos0 > 1.0f - EPSILON) {
			theta = 0.0f;
			// n is arbitrary
			n = new Vector(0.0f, 0.0f, 1.0f);
			
		} else { // cos0 < -1.0f + EPSILON must be true
			theta = (float)Math.PI;
			// determine the column of r + IDENTITY with the biggest length
			Matrix rPlusI = r.add(Matrix.IDENTITY);
			
			Vector c1 = rPlusI.getC1();
			Vector c2 = rPlusI.getC2();
			Vector c3 = rPlusI.getC3();
			
			n = c1;
			if (c2.length() > n.length()) {
				n = c2;
			}
			if (c3.length() > n.length()) {
				n = c3;
			}
			// we have to normalize n
			n = n.multiply(1.0f/n.length());
			
		}
		
		return new Rotation(n, theta);
	}
}
