/*
This class will be instantiated once and will contain all GUI elements of the
main window:
	- mainFrame, the top level window, to which will be added:
		- drawingPanel, to which will be added:
			- treeCanvas where our 3D drawing will be done;
		- controlPanel, to which will be added:
			- controlPanelLabel where a fixed text is displayed;
			- buttonPanel for screen control, to which will be added:
				- 12 buttons which can be used for movement;
		- menuBar, to which will be added:
			- fileMenu, to which will be added:
				- openMenuItem which can be used to read a tree file;
				- exitMenuItem which can be used to terminate the program;
			- animationMenu, to which will be added:
				- loadMenuItem which can be used to load an animation file;
				- saveMenuItem which can be used to save an animation file;
				- startMenuItem which can be used to start the animation;
				- controlMenuItem which can be used to toggle the Animation
				  control dialog on and off;
			- viewMenu, to which will be added:
				- resetViewpointMenuItem which can be used to reset the
				  viewpoint;
			- helpMenu, to which will be added:
				- aboutMenuItem which can be used to display some informative
				  text.
*/

package gui;

import java.awt.*;
import java.awt.event.*;
import java.util.HashMap;
import movement.Direction;
import movement.ScreenControlRunner;
import movement.ScreenControlMouseAdapter;
import tree.Tree;

import static display.Screen.MAIN_SCREEN;

public class MainWindow {
	
	
	// GUI elements:
	private Frame mainFrame;
	private Panel drawingPanel, controlPanel, buttonPanel;
	private Canvas treeCanvas;
	private Label controlPanelLabel;
	private Button xsp, xsn, ysp, ysn, zsp, zsn;
		// buttons that perform an x, y or z shift (s) in the positive (p) or
		// negative (n) direction
	private Button xryz, xrzy, yrxz, yrzx, zrxy, zryx;
		// buttons that perform an x, y or z rotate (r) from x, y or z to x, y
		// or z
	private MenuBar menuBar;
	private Menu fileMenu, animationMenu, viewMenu, helpMenu;
	private MenuItem openMenuItem, exitMenuItem, loadMenuItem, saveMenuItem,
	 startMenuItem, resetViewpointMenuItem, aboutMenuItem;
	private CheckboxMenuItem controlMenuItem;
	private MenuShortcut mSOpen;
	
	private ScreenControlRunner sCRunner;
	private ToggleControlItemListener tCItemListener;
	
	public MainWindow() {
		launchFrame();
	}
	
	
	public Canvas getTreeCanvas() {
		return treeCanvas;
	}
	
	public CheckboxMenuItem getControlMenuItem() {
		return controlMenuItem;
	}
	
	
	private void launchFrame() {
		mainFrame = new Frame("The Tree Project");
		mainFrame.setLayout(new FlowLayout());
		drawingPanel = new Panel();
		treeCanvas = new TreeCanvas(true); // manual mode for this canvas
		controlPanel = new Panel(new BorderLayout());
		controlPanelLabel = new Label("Screen control", Label.CENTER);
		buttonPanel = new Panel();
		buttonPanel.setLayout(new GridLayout(6,2));
		
		xsp = new Button("RIGHT");
		xsn = new Button("LEFT");
		ysp = new Button("UP");
		ysn = new Button("DOWN");
		zsp = new Button("ZOOM IN");
		zsn = new Button("ZOOM OUT");
		xryz = new Button("PUSH TOP");
		xrzy = new Button("PUSH BOTTOM");
		yrzx = new Button("PUSH LEFT");
		yrxz = new Button("PUSH RIGHT");
		zrxy = new Button("ROTATE LEFT");
		zryx = new Button("ROTATE RIGHT");
		
		// create a HashMap that maps screen control buttons to directions
		HashMap<Object, Direction> sCBD = new HashMap<Object, Direction>();
		sCBD.put(xsp, Direction.X_SHIFT_POSITIVE);
		sCBD.put(xsn, Direction.X_SHIFT_NEGATIVE);
		sCBD.put(ysp, Direction.Y_SHIFT_POSITIVE);
		sCBD.put(ysn, Direction.Y_SHIFT_NEGATIVE);
		sCBD.put(zsp, Direction.Z_SHIFT_POSITIVE);
		sCBD.put(zsn, Direction.Z_SHIFT_NEGATIVE);
		sCBD.put(xryz, Direction.X_ROTATE_Y_TO_Z);
		sCBD.put(xrzy, Direction.X_ROTATE_Z_TO_Y);
		sCBD.put(yrzx, Direction.Y_ROTATE_Z_TO_X);
		sCBD.put(yrxz, Direction.Y_ROTATE_X_TO_Z);
		sCBD.put(zrxy, Direction.Z_ROTATE_X_TO_Y);
		sCBD.put(zryx, Direction.Z_ROTATE_Y_TO_X);
		
		// create an anonymous SelectMarkerMouseAdapter object and register it
		// with the canvas
		treeCanvas.addMouseListener(new SelectMarkerMouseAdapter(this));
		
		// create the ScreenControlRunner object and give it all the
		// information it needs
		sCRunner = new ScreenControlRunner(this, sCBD);
		// create a ScreenControlMouseAdapter object and give it all the
		// information it needs
		ScreenControlMouseAdapter sCMouseAdapter =
		 new ScreenControlMouseAdapter(sCRunner);
		// then register this ScreenControlMouseAdapter object with the screen
		// control buttons
		xsp.addMouseListener(sCMouseAdapter);
		xsn.addMouseListener(sCMouseAdapter);
		ysp.addMouseListener(sCMouseAdapter);
		ysn.addMouseListener(sCMouseAdapter);
		zsp.addMouseListener(sCMouseAdapter);
		zsn.addMouseListener(sCMouseAdapter);
		xryz.addMouseListener(sCMouseAdapter);
		xrzy.addMouseListener(sCMouseAdapter);
		yrzx.addMouseListener(sCMouseAdapter);
		yrxz.addMouseListener(sCMouseAdapter);
		zrxy.addMouseListener(sCMouseAdapter);
		zryx.addMouseListener(sCMouseAdapter);
		
		// put all GUI components in their containers
		drawingPanel.add(treeCanvas);
		
		buttonPanel.add(xsn);
		buttonPanel.add(xsp);
		buttonPanel.add(zsp);
		buttonPanel.add(ysp);
		buttonPanel.add(zsn);
		buttonPanel.add(ysn);
		buttonPanel.add(yrzx);
		buttonPanel.add(yrxz);
		buttonPanel.add(zrxy);
		buttonPanel.add(xryz);
		buttonPanel.add(zryx);
		buttonPanel.add(xrzy);
		
		controlPanel.add(controlPanelLabel, BorderLayout.NORTH);
		controlPanel.add(buttonPanel, BorderLayout.CENTER);
		
		mainFrame.add(drawingPanel);
		mainFrame.add(controlPanel);
		
		
		// create the menu bar, menus and menu items
		menuBar = new MenuBar();
		
		fileMenu = new Menu("File");
		openMenuItem = new MenuItem("Open...");
		mSOpen = new MenuShortcut(KeyEvent.VK_O); // create shortcut Ctrl+O
		openMenuItem.setShortcut(mSOpen);		  // for the "Open..." menu item
		String exitString;
		if (System.getProperty("user.country").equals("NL")) {
			exitString = "Pleurt op"; // 010 humor
		} else {
			exitString = "Exit";
		}
		exitMenuItem = new MenuItem(exitString);
		
		animationMenu = new Menu("Animation");
		loadMenuItem = new MenuItem("Load...");
		saveMenuItem = new MenuItem("Save...");
		startMenuItem = new MenuItem("Start");
		controlMenuItem = new CheckboxMenuItem("Control");
		
		viewMenu = new Menu("View");
		resetViewpointMenuItem = new MenuItem("Reset viewpoint");
		
		helpMenu = new Menu("Help");
		aboutMenuItem = new MenuItem("About");
		
		// create a ReadTreeActionListener object
		ReadTreeActionListener rTActionListener =
		 new ReadTreeActionListener(mainFrame);
		// then register this ReadTreeActionListener object with the "Open..."
		// menu item
		openMenuItem.addActionListener(rTActionListener);
		
		// choosing 'Exit" will terminate the program
		exitMenuItem.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				System.exit(0);
			}
		});
		
		// create a LoadAnimationActionListener object
		LoadAnimationActionListener lAActionListener =
		 new LoadAnimationActionListener(mainFrame, this);
		// then register this LoadAnimationActionListener object with the
		// "Load..." menu item
		loadMenuItem.addActionListener(lAActionListener);
		
		// create a SaveAnimationActionListener object
		SaveAnimationActionListener sAActionListener =
		 new SaveAnimationActionListener(mainFrame);
		// then register this SaveAnimationActionListener object with the
		// "Save..." menu item
		saveMenuItem.addActionListener(sAActionListener);
		
		// register an ActionListener object with the "Start" menu item
		startMenuItem.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				Tree.startAnimation();
			}
		});
		
		// create a ToggleControlItemListener object
		tCItemListener = new ToggleControlItemListener(this);
		// then register this ToggleControlItemListener object with the
		// "Control" menu item
		controlMenuItem.addItemListener(tCItemListener);
		
		// choosing 'Reset viewpoint' will reset the viewpoint of the current
		// tree
		resetViewpointMenuItem.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				// reset the viewpoint to the initial viewpoint
				MAIN_SCREEN.resetViewpoint();
				// find the screen representation
				MAIN_SCREEN.findRepresentation();
				// repaint
				treeCanvas.repaint();
			}
		});
		
		// choosing 'About' will launch a dialog
		aboutMenuItem.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				new AboutDialog(mainFrame);
			}
		});
		
		// assemble the memu structure
		fileMenu.add(openMenuItem);
		fileMenu.addSeparator();
		fileMenu.add(exitMenuItem);
		
		animationMenu.add(loadMenuItem);
		animationMenu.add(saveMenuItem);
		animationMenu.add(startMenuItem);
		animationMenu.add(controlMenuItem);
		
		viewMenu.add(resetViewpointMenuItem);
		
		helpMenu.add(aboutMenuItem);
		
		menuBar.add(fileMenu);
		
		menuBar.add(animationMenu);
		
		menuBar.add(viewMenu);
		
		menuBar.setHelpMenu(helpMenu);
		
		mainFrame.setMenuBar(menuBar);
		
		
		// enable the close window button
		mainFrame.addWindowListener(new WindowAdapter() {
			public void windowClosing(WindowEvent e) {
				System.exit(0);
			}
		});
		
		// finally, pack and show
		mainFrame.pack();
		mainFrame.setVisible(true);
		
		
	}
	
	
	// method doAnimation() is called from the main() method of the Tree class
	// to run a short animation
	// it just dispatches the call to the ScreenControlRunner object because
	// that is where all screen movement is done
	public void doAnimation() {
		sCRunner.doAnimation();
	}
	
	// method showDeleteErrorDialog() is called from the deleteMarker() method
	// of the Markers class in case the Delete button was clicked but no marker
	// was selected
	public void showDeleteErrorDialog() {
		new DeleteErrorDialog(tCItemListener);
	}
	
}
