/*
The tree canvas is where the 3D drawing will be done.
Note that the actual drawing is done in the paint() method.
The paint() method is never called directly but it is invoked when
- the tree canvas is made visible for the first time;
- the screen has moved to a new position in the run() method of the
  movement.ScreenControlRunner or movement.AnimationRunner class;
- the Java runtime environment deems it necessary.

In the first two cases the calculations for the screen representation of every
surface, branch, marker and trail segment must have been done otherwise the
paint() method will not have the desired effect.
*/

package gui;

import java.util.concurrent.locks.Lock;
import java.util.HashSet;
import display.OpaquePolygon;
import display.Screen;
import tree.Branch;
import tree.Branches;
import tree.Tree;
import visual_aid.Marker;
import visual_aid.Markers;
import visual_aid.Trail;
import visual_aid.TrailSegment;
import java.awt.*;

public class TreeCanvas extends Canvas {
	private Image offscreenImage;
	private Graphics2D offscreenG2;
	private final boolean manualMode;
	
	public TreeCanvas(boolean manualMode) {
		this.manualMode = manualMode;
		
		// setting the background color upon construction eliminates an initial
		// one time flicker
		setBackground(new Color(Tree.BG_COLOR)); // background has color
												 // determined by constant
		
		// using setPreferredSize() does the right thing whereas setSize() does
		// not
		setPreferredSize(new Dimension(Screen.SCREEN_WIDTH, Screen.SCREEN_HEIGHT));
	}
	
	public void update(Graphics g) {
		// the update() method overrides and deprecates the standard update()
		// method: instead of clearing the component and calling the paint()
		// method, now we are just calling the paint() method
		paint(g);
	}
	
	public void paint(Graphics g) {
		// the paint() method does all the drawing to the graphic context of an
		// offscreen image, and then draws this offscreen image to the normal
		// graphic context
		
		// the paint() method concerns itself with drawing all surfaces,
		// branches, markers and trail segments but it does not concern itself
		// with calculating their screen representations, it assumes that all
		// these calculations have been done
		
		// note that in the following code, first the surfaces are drawn, then
		// the branches, then the trail segments and lastly the markers, this
		// causes surfaces to be covered by branches, branches by trail
		// segments and trail segments by markers, which is what we want
		
		boolean drawTheRest;
		
		offscreenImage = createImage(Screen.SCREEN_WIDTH, Screen.SCREEN_HEIGHT);
		offscreenG2 = (Graphics2D)offscreenImage.getGraphics();
		 // here we obtain the Graphics2D ojbect associated with the offscreen
		 // image
		offscreenG2.setColor(new Color(Tree.BG_COLOR));
		 // background has color determined by constant
		offscreenG2.fillRect(0, 0, Screen.SCREEN_WIDTH, Screen.SCREEN_HEIGHT); 
		
		Lock lock = manualMode ?
		 Screen.MAIN_SCREEN.getOPLock() : Screen.ANIMATION_SCREEN.getOPLock();
		if (lock.tryLock()) { // if not locked by other trhead, set the lock
			try {
				drawTheRest = true; // also draw the rest
									// i.e. branches, trail semgents, markers
				
				// draw surfaces
				HashSet<OpaquePolygon> opaquePolygons = manualMode ?
				 Screen.MAIN_SCREEN.getOpaquePolygons() : Screen.ANIMATION_SCREEN.getOpaquePolygons();
				if (opaquePolygons != null) {
					for (OpaquePolygon opaquePolygon : opaquePolygons) {
						int[] xPoints = opaquePolygon.getXPoints();
						int[] yPoints = opaquePolygon.getYPoints();
						
						Color c = opaquePolygon.getColor();
						offscreenG2.setColor(c);
						
						// draw the surface
						offscreenG2.fillPolygon(xPoints, yPoints, xPoints.length);
					}
				}
			} finally {
				lock.unlock(); // whatever happens, always unlock (standard mechanism)
			}
		} else {
			drawTheRest = false; // surfaces were not drawn, so neither are
								 // branches, trail semgents, markers
		}
		
		if (drawTheRest) {
			// draw branches
			Branches branches = Branches.getInstance();
			for (Branch branch : branches.values()) {
				if (branch.getHasScreenRepresentation()) {
					offscreenG2.setStroke(new BasicStroke(branch.getScreenWidth(),
					 BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND));
					 // pen has round endings as well as round joins (not used)
					
					// the color of the branch is a mixture of the pen and
					// background colors
					float partOfPenColor = branch.getScreenPartOfPenColor();
					float partOfBgColor = 1.0f - partOfPenColor;
					
					Color c = new Color((int)(partOfPenColor * Tree.PEN_RED + partOfBgColor * Tree.BG_RED),
					 (int)(partOfPenColor * Tree.PEN_GREEN + partOfBgColor * Tree.BG_GREEN),
					 (int)(partOfPenColor * Tree.PEN_BLUE + partOfBgColor * Tree.BG_BLUE));
					offscreenG2.setColor(c);
					
					// draw the branch
					offscreenG2.drawLine(branch.getX1(), branch.getY1(),
					 branch.getX2(), branch.getY2());
				}
			}
			
			// draw trail segments
			boolean showTrail = manualMode ? Trail.showTrailInManualMode : Trail.showTrailInAnimationMode;
			if (showTrail) {
				offscreenG2.setStroke(new BasicStroke(Trail.TRAIL_SEGMENT_SIZE,
				 BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND));
				 // pen has round endings as well as round joins (not used)
				
				offscreenG2.setColor(new Color(Markers.MARKER_COLOR));
				 // the color of the trail segments is the marker color
				
				Trail trail = Trail.getInstance();
				for (TrailSegment trailSegment : trail) {
					if (trailSegment.getHasScreenRepresentation()) {
						// draw the trail segment
						offscreenG2.drawLine(trailSegment.getX1(), trailSegment.getY1(),
						 trailSegment.getX2(), trailSegment.getY2());
					}
				}
			}
			
			// draw markers
			boolean showMarkers = manualMode ? Markers.showMarkersInManualMode : Markers.showMarkersInAnimationMode;
			if (showMarkers) {
				Markers markers = Markers.getInstance();
				
				// RadialGradientPaint is used to give a smooth appearance
				
				// set the float[] array for the colors of the
				// RadialGradientPaint, the mark up to which the pen color will
				// be opaque is set by the first value, then the color will
				// gradually become more transparent and will have become
				// completely transparent at the circumference of the disc
				float[] fractions = {Markers.FIRST_KEYFRAME, 1.0f};
				
				for (Marker marker : markers) {
					if (marker.getHasScreenRepresentation()) {
						// depending on whether this is an unselected marker or
						// a selected marker (can only happen once), we create
						// a different Color[] array for RadialGradientPaint
						Color[] colors;
						if (marker != Markers.selectedMarker) {
							colors = new Color[]
							 {new Color(Markers.MARKER_COLOR, false),
							 new Color(Markers.MARKER_COLOR, true)};
						} else {
							colors = new Color[]
							 {new Color(Markers.SELECTED_MARKER_COLOR, false),
							 new Color(Markers.SELECTED_MARKER_COLOR, true)};
						}
						
						// now we can create the RadialGraintPaint object
						RadialGradientPaint rgp = new RadialGradientPaint(
						 marker.getX1(), marker.getY1(),
						 marker.getScreenSize(), fractions, colors);
						offscreenG2.setPaint(rgp);
						
						// draw the marker
						// note that we add 1 to the size of the bounding box,
						// this makes for a better appearance
						offscreenG2.fillOval(Math.round(marker.getX1() - marker.getScreenSize()),
						 Math.round(marker.getY1() - marker.getScreenSize()),
						 Math.round(2.0f * marker.getScreenSize()) + 1,
						 Math.round(2.0f * marker.getScreenSize()) + 1);
					}
				}
			}
			
			g.drawImage(offscreenImage, 0, 0, this);
			
		}
		
	}
	
}
