/*
An AnimationRunner object will run a separate thread as long as the animation
is not stopped in any way by the user. This thread performs the actual
animation.
*/

package movement;

import java.util.Iterator;

import geometry.Matrix;
import geometry.Rotation;
import geometry.Vector;
import gui.AnimationWindow;
import visual_aid.Trail;
import visual_aid.TrailSegment;

import static display.Screen.ANIMATION_SCREEN;
import static display.Screen.RETINA_SCREEN_DISTANCE;

public class AnimationRunner implements Runnable {
	
	private final int SLEEP_TIME = 10;
	private int sleepTime = SLEEP_TIME;
	
	private final float SHIFT = 10.0f;
	private float shift = SHIFT;
	
	
	private AnimationWindow animationWindow;
	
	public AnimationRunner(AnimationWindow animationWindow) {
		this.animationWindow = animationWindow; // tell the instance object which window to use
	}
	
	public void run() {
		Trail trail = Trail.getInstance();
		Iterator<TrailSegment> trailIterator = trail.iterator();
		
		if (trailIterator.hasNext()) { // only consider real trails, i.e. at least one segment
			
			float distanceFromStartSegment = 0.0f; // keep track of the distance
												   // to the start of the segment
			boolean startOfTrail = true; // we always have an animation step
										 // for the start of the trail
										 // even if the first segment has length 0
			TrailSegment trailSegment;
			do {
				trailSegment = trailIterator.next();
				// determine the length of the segment and the rotational
				// distance from start marker to end marker
				float segmentLength =
				 trailSegment.getEndMarker().subtract(trailSegment.getStartMarker()).length();
				Matrix startBase = new Matrix(trailSegment.getStartMarker().getU(),
											  trailSegment.getStartMarker().getV(),
											  trailSegment.getStartMarker().getNormal());
				Matrix endBase   = new Matrix(trailSegment.getEndMarker().getU(),
											  trailSegment.getEndMarker().getV(),
											  trailSegment.getEndMarker().getNormal());
				Matrix rotationalDistance = endBase.multiply(startBase.transpose());
				
				// now calculate the axis of rotaton and the rotation angle
				// when moving from start marker to end marker
				Rotation rotation = Rotation.convertFromMatrix(rotationalDistance);
				Vector n = rotation.getN();
				float theta = rotation.getTheta();
				
				// do an animation step as long as there is some segment part left
				// (or if we are at the start of the trail)
				while (distanceFromStartSegment < segmentLength || startOfTrail) {
					
					if (!startOfTrail) {
						// first: calculate the new screen middle and the base
						float relativeDistance = distanceFromStartSegment/segmentLength;
						
						ANIMATION_SCREEN.screenMiddle =
						 trailSegment.getStartMarker().add(trailSegment.getEndMarker().subtract(trailSegment.getStartMarker()).multiply(relativeDistance));
						float relativeTheta = relativeDistance * theta;
						
						rotation = new Rotation(n, relativeTheta); // (reuse)
						Matrix base = rotation.convertToMatrix().multiply(startBase);
						
						ANIMATION_SCREEN.u      = base.getC1();
						ANIMATION_SCREEN.v      = base.getC2();
						ANIMATION_SCREEN.normal = base.getC3();
						
					} else { // the start of the trail is a special situation
						// first set the new screen middle and the base
						ANIMATION_SCREEN.screenMiddle =
						 trailSegment.getStartMarker();
						ANIMATION_SCREEN.u      = trailSegment.getStartMarker().getU();
						ANIMATION_SCREEN.v      = trailSegment.getStartMarker().getV();
						ANIMATION_SCREEN.normal = trailSegment.getStartMarker().getNormal();
						startOfTrail = false;
						
					}
					
					// second: calculate the retina
					ANIMATION_SCREEN.retina =
					 ANIMATION_SCREEN.screenMiddle.subtract(ANIMATION_SCREEN.normal.multiply(RETINA_SCREEN_DISTANCE));
					
					// third: find the screen represenation
					ANIMATION_SCREEN.findRepresentation();
					
					// fourth: repaint
					animationWindow.getTreeCanvas().repaint();
					
					// fifth: sleep to slow down
					try {
						Thread.sleep(sleepTime);
					} catch (InterruptedException e) {
						System.err.println("WARNING: thread had its sleep interrupted in AnimationRunner");
					}
					
					// sixth: prepare for the next animation step
					distanceFromStartSegment += shift;
					
				}
				
				// once there is no segment part left move to the next segment,
				// first recalculate the distance from the start of the segment
				distanceFromStartSegment -= segmentLength;
				
			} while (trailIterator.hasNext());
			
			// finally, always do an animation step for the end of the trail
			ANIMATION_SCREEN.screenMiddle = trailSegment.getEndMarker();
			ANIMATION_SCREEN.u      = trailSegment.getEndMarker().getU();
			ANIMATION_SCREEN.v      = trailSegment.getEndMarker().getV();
			ANIMATION_SCREEN.normal = trailSegment.getEndMarker().getNormal();
			ANIMATION_SCREEN.retina =
			 ANIMATION_SCREEN.screenMiddle.subtract(ANIMATION_SCREEN.normal.multiply(RETINA_SCREEN_DISTANCE));
			ANIMATION_SCREEN.findRepresentation();
			animationWindow.getTreeCanvas().repaint();
			
		} else {
			// reserved for info message
		}
		
	}
	
}
