/*
A tree is constructed of nodes and branches.
A node is a point somewhere in the three dimensional space.
A branch is a straight line between two nodes.

A branch has a visible part. The visible part depends on which
method we use: the screen plane method or the retina plane method.
The visible part has end points p' and q' but a method
calculateScreenPlaneVisibility() or calculateRetinaPlaneVisibility()
must be called to calculate these.
*/

package tree;

import display.Screen;
import display.Visibility;
import geometry.Line;
import geometry.Matrix;
import geometry.Vector;

public class Branch extends Line {
	
	private Node node1, node2;
	
	private Vector pPrime, qPrime;
	// end points for the visible part of the branch
	//
	// for the screen plane method:
	// if pPrime and qPrime have null values after this calculation,
	// then p respectively q was on the same side of the screen plane as the retina plane
	// and q respectively p was on the screen plane, or
	// both p and q were on the same side of the screen plane as the retina plane
	//
	// for the retina plane method:
	// if pPrime and qPrime have null values after this calculation,
	// then p was on the retina plane or on the other side of the retina plane as the screen plane, and
	// q was on the retina plane or on the other side of the retina plane as the screen plane
	
	private boolean pPrimeOnPlane, qPrimeOnPlane;
	// for the screen plane method:
	// when projecting to the screen it will save time to know if an end point
	// is on the screen plane itself
	//
	// for the retina plane method:
	// when projecting to the screen it is essential to know if an end point
	// is on the retina plane itself
	
	private Vector p2Prime, q2Prime;
	// end points for the projection of the visible part of the branch on the screen plane
	//
	// for the screen plane method:
	// if pPrime and qPrime do not have null values then projection can be done
	// note that if pPrimeOnPlane equals true then no calculations are needed for p2Prime,
	// same for qPrimeOnPlane
	// if pPrime and qPrime have null values then p2Prime and q2Prime will get null values too
	//
	// for the retina plane method:
	// if pPrime and qPrime do not have null values then projection can be done
	// note that if pPrime is on the retina plane then p2Prime will be set to pPrime
	// because we will use the vector pPrime - retina to determine the halfline stretching from
	// q2Prime into infinity in the planar representation
	// same for qPrime
	// if pPrime and qPrime have null values then p2Prime and q2Prime will get null values too
	
	private float ax, ay, bx, by;
	// 2D representation of the projection of the visible part of the branch
	// in terms of u and v
	//
	// for the screen plane method:
	// if both p2Prime and q2Prime are unequal to null, then
	// p2Prime - screenMiddle = ax * u + ay * v and
	// q2Prime - screenMiddle = bx * u + by * v
	//
	// for the retina plane method:
	// if both p2Prime and q2Prime are unequal to null, then
	//
	// if both pPrimeOnPlane and qPrimeOnPlane are false then
	// p2Prime - screenMiddle = ax * u + ay * v and
	// q2Prime - screenMiddle = bx * u + by * v
	//
	// if pPrimeOnPlane is false and qPrimeOnPlane is true, then
	// p2Prime - screenMiddle = ax * u + ay * v and
	// q2Prime - retina       = bx * u + by * v
	//  (note that then q2Prime equals qPrime and is on the retina plane)
	//
	// if pPrimeOnPlane is true and qPrimeOnPlane is false, then
	// q2Prime - screenMiddle = ax * u + ay * v and
	// p2Prime - retina       = bx * u + by * v
	//  (note that then p2Prime equals pPrime and is on the retina plane)
	
	private float axPrime, ayPrime, bxPrime, byPrime;
	// 2D representation of the intersection of the projection of the visible part of the branch
	// with the screen in terms of u and v
	
	private float screenWidth;
	// the width of the branch on the screen taking into account the
	// enhancement effect
	
	private float screenPartOfPenColor;
	// the part of the pen color in the color of the branch on the screen
	// taking into account the enhancement effect
	
	private boolean hasScreenRepresentation;
	// indicates if the branch is (partly) visible or not
	
	private int x1, y1, x2, y2;
	// (partly) visible branch in terms of pixel coordinates
	// 0 <= .. < SCREEN_WIDTH for x1 and x2 and 0 <= .. < SCREEN_HEIGHT for y1 and y2
	
	
	// The constructor creates a branch that references two nodes.
	public Branch(Node node1, Node node2) {
		super(node1, node2);
		this.node1 = node1;
		this.node2 = node2;
	}
	
	public Node getNode1() {
		return node1;
	}
	
	public Node getNode2() {
		return node2;
	}
	
	public Vector getPPrime() {
		return pPrime;
	}
	
	public Vector getQPrime() {
		return qPrime;
	}
	
	public boolean getPPrimeOnPlane() {
		return pPrimeOnPlane;
	}
	
	public boolean getQPrimeOnPlane() {
		return qPrimeOnPlane;
	}
	
	public Vector getP2Prime() {
		return p2Prime;
	}
	
	public Vector getQ2Prime() {
		return q2Prime;
	}
	
	public float getAx() {
		return ax;
	}
	
	public float getAy() {
		return ay;
	}
	
	public float getBx() {
		return bx;
	}
	
	public float getBy() {
		return by;
	}
	
	public float getAxPrime() {
		return axPrime;
	}
	
	public float getAyPrime() {
		return ayPrime;
	}
	
	public float getBxPrime() {
		return bxPrime;
	}
	
	public float getByPrime() {
		return byPrime;
	}
	
	public float getScreenWidth() {
		return screenWidth;
	}
	
	public float getScreenPartOfPenColor() {
		return screenPartOfPenColor;
	}
	
	public boolean getHasScreenRepresentation() {
		return hasScreenRepresentation;
	}
	
	public int getX1() {
		return x1;
	}
	
	public int getY1() {
		return y1;
	}
	
	public int getX2() {
		return x2;
	}
	
	public int getY2() {
		return y2;
	}
	
	
	
	// method for calculating the visible part of the line
	// for the screen plane method
	// this method calculates pPrime, qPrime, pPrimeOnPlane
	// and qPrimeOnPlane
	/* public int calculateVisibilitySPM() {
		int a = 0;
		Visibility pVisibility = Screen.visibilitySPM(p);
		Visibility qVisibility = Screen.visibilitySPM(q);
		
		switch (pVisibility) {
		case VISIBLE:
			switch (qVisibility) {
			case VISIBLE:
				pPrime = p;
				qPrime = q;
				pPrimeOnPlane = false;
				qPrimeOnPlane = false;
				a = 1;
				break;
			case ON_PLANE:
				pPrime = p;
				qPrime = q;
				pPrimeOnPlane = false;
				qPrimeOnPlane = true;
				a = 2;
				break;
			case INVISIBLE:
				// see the Screen class
				// we are determining lambda such that
				// p + lambda (q - p) = s^m + mu u + nu v
				// or [ p - q  u  v ] (lambda, mu, nu)^T = p - s^m
				// we use Cramer's rule to calculate lambda
				Matrix M = new Matrix(p.subtract(q), Screen.u, Screen.v);
				Matrix M_lambda = new Matrix(p.subtract(Screen.screenMiddle),
					Screen.u, Screen.v);
				float lambda = M_lambda.determinant() / M.determinant();
				pPrime = p;
				qPrime = p.add(q.subtract(p).multiply(lambda));
				pPrimeOnPlane = false;
				qPrimeOnPlane = true;
				a =3;
				break;
			}
			break;
		case ON_PLANE:
			switch (qVisibility) {
			case VISIBLE:
				pPrime = p;
				qPrime = q;
				pPrimeOnPlane = true;
				qPrimeOnPlane = false;
				a = 4;
				break;
			case ON_PLANE:
				pPrime = p;
				qPrime = q;
				pPrimeOnPlane = true;
				qPrimeOnPlane = true;
				a = 5;
				break;
			case INVISIBLE:
				pPrime = null;
				qPrime = null;
				pPrimeOnPlane = true;
				qPrimeOnPlane = false;
				a = 6;
				break;
			}
			break;
		case INVISIBLE:
			switch (qVisibility) {
			case VISIBLE:
				// see the Screen class
				// we are determining lambda such that
				// p + lambda (q - p) = s^m + mu u + nu v
				// or [ p - q  u  v ] (lambda, mu, nu)^T = p - s^m
				// we use Cramer's rule to calculate lambda
				Matrix M = new Matrix(p.subtract(q), Screen.u, Screen.v);
				Matrix M_lambda = new Matrix(p.subtract(Screen.screenMiddle),
					Screen.u, Screen.v);
				float lambda = M_lambda.determinant() / M.determinant();
				pPrime = p.add(q.subtract(p).multiply(lambda));
				qPrime = q;
				pPrimeOnPlane = true;
				qPrimeOnPlane = false;
				a = 7;
				break;
			case ON_PLANE:
				pPrime = null;
				qPrime = null;
				pPrimeOnPlane = false;
				qPrimeOnPlane = true;
				a = 8;
				break;
			case INVISIBLE:
				pPrime = null;
				qPrime = null;
				pPrimeOnPlane = false;
				qPrimeOnPlane = false;
				a = 9;
				break;
			}
			break;
		}
		return a;
	} */
	
	// method for calculating the visible part of the line
	// for the retina plane method
	// this method calculates pPrime, qPrime, pPrimeOnPlane
	// and qPrimeOnPlane
	public int calculateVisibilityRPM(Screen screen) {
		int a = 0;
		Visibility pVisibility = screen.visibilityRPM(p);
		Visibility qVisibility = screen.visibilityRPM(q);
		
		switch (pVisibility) {
		case VISIBLE:
			switch (qVisibility) {
			case VISIBLE:
				pPrime = p;
				qPrime = q;
				pPrimeOnPlane = false;
				qPrimeOnPlane = false;
				a = 1;
				break;
			case ON_PLANE:
				pPrime = p;
				qPrime = q;
				pPrimeOnPlane = false;
				qPrimeOnPlane = true;
				a = 2;
				break;
			case INVISIBLE:
				// see the Screen class
				// we are determining lambda such that
				// p + lambda (q - p) = r + mu u + nu v
				// or [ p - q  u  v ] (lambda, mu, nu)^T = p - r
				// we use Cramer's rule to calculate lambda
				Matrix M = new Matrix(p.subtract(q), screen.u, screen.v);
				Matrix M_lambda = new Matrix(p.subtract(screen.retina),
					screen.u, screen.v);
				float lambda = M_lambda.determinant() / M.determinant();
				pPrime = p;
				qPrime = p.add(q.subtract(p).multiply(lambda));
				pPrimeOnPlane = false;
				qPrimeOnPlane = true;
				a =3;
				break;
			}
			break;
		case ON_PLANE:
			switch (qVisibility) {
			case VISIBLE:
				pPrime = p;
				qPrime = q;
				pPrimeOnPlane = true;
				qPrimeOnPlane = false;
				a = 4;
				break;
			case ON_PLANE:
				pPrime = null;
				qPrime = null;
				pPrimeOnPlane = true;
				qPrimeOnPlane = true;
				a = 5;
				break;
			case INVISIBLE:
				pPrime = null;
				qPrime = null;
				pPrimeOnPlane = true;
				qPrimeOnPlane = false;
				a = 6;
				break;
			}
			break;
		case INVISIBLE:
			switch (qVisibility) {
			case VISIBLE:
				// see the Screen class
				// we are determining lambda such that
				// p + lambda (q - p) = r + mu u + nu v
				// or [ p - q  u  v ] (lambda, mu, nu)^T = p - r
				// we use Cramer's rule to calculate lambda
				Matrix M = new Matrix(p.subtract(q), screen.u, screen.v);
				Matrix M_lambda = new Matrix(p.subtract(screen.retina),
					screen.u, screen.v);
				float lambda = M_lambda.determinant() / M.determinant();
				pPrime = p.add(q.subtract(p).multiply(lambda));
				qPrime = q;
				pPrimeOnPlane = true;
				qPrimeOnPlane = false;
				a = 7;
				break;
			case ON_PLANE:
				pPrime = null;
				qPrime = null;
				pPrimeOnPlane = false;
				qPrimeOnPlane = true;
				a = 8;
				break;
			case INVISIBLE:
				pPrime = null;
				qPrime = null;
				pPrimeOnPlane = false;
				qPrimeOnPlane = false;
				a = 9;
				break;
			}
			break;
		}
		return a;
	}
	
	// method for projecting the line on the screen
	// for the screen plane method
	// this method calculates p2Prime and q2Prime
	/* public void projectOnScreenSPM() {
		if (pPrime != null & qPrime != null) {
			if (!pPrimeOnPlane) {
				// see the Screen class
				// we are determining lambda such that
				// p' + lambda (r - p') = s^m + mu u + nu v
				// or [ p' - r  u  v ] (lambda, mu, nu)^T = p' - s^m
				// we use Cramer's rule to calculate lambda
				Matrix M = new Matrix(pPrime.subtract(Screen.retina),
					Screen.u, Screen.v);
				Matrix M_lambda = new Matrix(pPrime.subtract(Screen.screenMiddle),
					Screen.u, Screen.v);
				float lambda = M_lambda.determinant() / M.determinant();
				p2Prime = pPrime.add(Screen.retina.subtract(pPrime).multiply(lambda));
			} else {
				p2Prime = pPrime;
			}
			if (!qPrimeOnPlane) {
				// see explanation for determining p2Prime
				Matrix M = new Matrix(qPrime.subtract(Screen.retina),
					Screen.u, Screen.v);
				Matrix M_lambda = new Matrix(qPrime.subtract(Screen.screenMiddle),
					Screen.u, Screen.v);
				float lambda = M_lambda.determinant() / M.determinant();
				q2Prime = qPrime.add(Screen.retina.subtract(qPrime).multiply(lambda));
			} else {
				q2Prime = qPrime;
			}
		} else {
			// if pPrime and qPrime have null values then we set
			// p2Prime and qPrime to null values as well
			p2Prime = null;
			q2Prime = null;
		}
		
	} */
	
	// method for projecting the line on the screen
	// for the retina plane method
	// this method calculates p2Prime and q2Prime
	public void projectOnScreenRPM(Screen screen) {
		if (pPrime != null & qPrime != null) {
			if (!pPrimeOnPlane) {
				// see the Screen class
				// we are determining lambda such that
				// p' + lambda (r - p') = s^m + mu u + nu v
				// or [ p' - r  u  v ] (lambda, mu, nu)^T = p' - s^m
				// we use Cramer's rule to calculate lambda
				Matrix M = new Matrix(pPrime.subtract(screen.retina),
					screen.u, screen.v);
				Matrix M_lambda = new Matrix(pPrime.subtract(screen.screenMiddle),
					screen.u, screen.v);
				float lambda = M_lambda.determinant() / M.determinant();
				p2Prime = pPrime.add(screen.retina.subtract(pPrime).multiply(lambda));
			} else {
				p2Prime = pPrime;
			}
			if (!qPrimeOnPlane) {
				// see explanation for determining p2Prime
				Matrix M = new Matrix(qPrime.subtract(screen.retina),
					screen.u, screen.v);
				Matrix M_lambda = new Matrix(qPrime.subtract(screen.screenMiddle),
					screen.u, screen.v);
				float lambda = M_lambda.determinant() / M.determinant();
				q2Prime = qPrime.add(screen.retina.subtract(qPrime).multiply(lambda));
			} else {
				q2Prime = qPrime;
			}
		} else {
			// if pPrime and qPrime have null values then we set
			// p2Prime and qPrime to null values as well
			p2Prime = null;
			q2Prime = null;
		}
		
	}
	
	// method for calculating the 2D planar representation
	// for the screen plane method
	// this method calculates ax, ay, bx and by
	/* public boolean calculatePlanarRepresentationSPM() {
		// returns true if this line has a valid planar representation,
		// returns false otherwise
		if (p2Prime != null & q2Prime != null) {
			ax = p2Prime.subtract(Screen.screenMiddle).innerProduct(Screen.u);
			ay = p2Prime.subtract(Screen.screenMiddle).innerProduct(Screen.v);
			bx = q2Prime.subtract(Screen.screenMiddle).innerProduct(Screen.u);
			by = q2Prime.subtract(Screen.screenMiddle).innerProduct(Screen.v);
			return true;
		} else {
			return false;
		}
	} */
	
	// method for calculating the 2D planar representation
	// for the retina plane method
	// this method calculates ax, ay, bx and by
	public boolean calculatePlanarRepresentationRPM(Screen screen) {
		// returns true if this line has a valid planar representation,
		// returns false otherwise
		if (p2Prime != null & q2Prime != null) {
			if (!pPrimeOnPlane & !qPrimeOnPlane) {
				ax = p2Prime.subtract(screen.screenMiddle).innerProduct(screen.u);
				ay = p2Prime.subtract(screen.screenMiddle).innerProduct(screen.v);
				bx = q2Prime.subtract(screen.screenMiddle).innerProduct(screen.u);
				by = q2Prime.subtract(screen.screenMiddle).innerProduct(screen.v);
				return true;
			} else if (!pPrimeOnPlane & qPrimeOnPlane) {
				ax = p2Prime.subtract(screen.screenMiddle).innerProduct(screen.u);
				ay = p2Prime.subtract(screen.screenMiddle).innerProduct(screen.v);
				bx = q2Prime.subtract(screen.retina).innerProduct(screen.u);
				by = q2Prime.subtract(screen.retina).innerProduct(screen.v);
				return true;
			} else if (pPrimeOnPlane & !qPrimeOnPlane) {
				ax = q2Prime.subtract(screen.screenMiddle).innerProduct(screen.u);
				ay = q2Prime.subtract(screen.screenMiddle).innerProduct(screen.v);
				bx = p2Prime.subtract(screen.retina).innerProduct(screen.u);
				by = p2Prime.subtract(screen.retina).innerProduct(screen.v);
				return true;
			} else { // this case should never happen
				return false;
			}
		} else {
			return false;
		}
	}
	
	
	// method for calculating the intersection
	// of the 2D planar representation with the screen
	// for the screen plane method
	// returns true if the representation has an intersection with the screen
	// returns false otherwise
	// pre-condition: method calculatePlanarRepresentationSPM() returned true
	/* public boolean findIntersectionSPM() {
		
		float W = Screen.SCREEN_WIDTH/2.0f;
		float H = Screen.SCREEN_HEIGHT/2.0f;
		float lambdaA, lambdaB, maxArgX, maxArgY, minArgX, minArgY; */
		
		/* we search those lambda such that:
			-W <= ax + lambda * (bx - ax)
			ax + lambda * (bx - ax) <= W
			-H <= ay + lambda * (by - ay)
			ay + lambda * (by - ay) <= H
			0 <= lambda
			lambda <= 1
		   if, for example, bx - ax > 0 and by - ay > 0:
			lambdaA = max { (-W - ax) / (bx - ax), (-H - ay) / (by - ay), 0 }
			lambdaB = min { (W - ax) / (bx - ax), (H - ay) / (by - ay), 1 }
		   now if lambdaA <= lambdaB we have some part of the line that
		   intersects with the screen
		*/
		/* if (bx > ax) {
			maxArgX = (-W-ax)/(bx-ax);
			minArgX = (W-ax)/(bx-ax);
		} else if (bx < ax) {
			maxArgX = (W-ax)/(bx-ax);
			minArgX = (-W-ax)/(bx-ax);
		} else { // ax = bx
			if ((ax <= W) & (ax >= -W)) {
				maxArgX = 0.0f; // x coordinate is in bounds of screen
				minArgX = 1.0f; // so let it have no influence on lambdA and lambdaB
			} else {
				maxArgX = 2.0f; // line is outside screen
				minArgX = -1.0f; // so force lambdaA > lambdaB
			}
		}
		
		if (by > ay) {
			maxArgY = (-H-ay)/(by-ay);
			minArgY = (H-ay)/(by-ay);
		} else if (by < ay) {
			maxArgY = (H-ay)/(by-ay);
			minArgY = (-H-ay)/(by-ay);
		} else { // ay = by
			if ((ay <= H) & (ay >= -H)) {
				maxArgY = 0.0f; // y coordinate is in bounds of screen
				minArgY = 1.0f; // so let it have no influence on lambdA and lambdaB
			} else {
				maxArgY = 2.0f; // line is outside screen
				minArgY = -1.0f; // so force lambdaA > lambdaB
			}
		}
		
		lambdaA = Math.max(maxArgX, maxArgY);
		lambdaA = Math.max(lambdaA, 0.0f);
		lambdaB = Math.min(minArgX, minArgY);
		lambdaB = Math.min(lambdaB, 1.0f);
		
		if (lambdaA <= lambdaB) {
			axPrime = ax + lambdaA * (bx - ax);
			ayPrime = ay + lambdaA * (by - ay);
			bxPrime = ax + lambdaB * (bx - ax);
			byPrime = ay + lambdaB * (by - ay);
			return true;
		} else {
			return false;
		}
		
	} */
	
	
	// method for calculating the intersection
	// of the 2D planar representation with the screen
	// for the retina plane method
	// returns true if the representation has an intersection with the screen
	// returns false otherwise
	// pre-condition: method calculatePlanarRepresentationRPM() returned true
	public boolean findIntersectionRPM() {
		
		float W = Screen.SCREEN_WIDTH/2.0f;
		float H = Screen.SCREEN_HEIGHT/2.0f;
		float lambdaA, lambdaB, maxArgX, maxArgY, minArgX, minArgY;
		boolean minArgXInfinity, minArgYInfinity;
		
		if (!pPrimeOnPlane & !qPrimeOnPlane) { // planar representation is bounded
		/* we search those lambda such that:
			-W <= ax + lambda * (bx - ax)
			ax + lambda * (bx - ax) <= W
			-H <= ay + lambda * (by - ay)
			ay + lambda * (by - ay) <= H
			0 <= lambda
			lambda <= 1
		   if, for example, bx - ax > 0 and by - ay > 0:
			lambdaA = max { (-W - ax) / (bx - ax), (-H - ay) / (by - ay), 0 }
			lambdaB = min { (W - ax) / (bx - ax), (H - ay) / (by - ay), 1 }
		   now if lambdaA <= lambdaB we have some part of the line that
		   intersects with the screen
		*/
			if (bx > ax) {
				maxArgX = (-W-ax)/(bx-ax);
				minArgX = (W-ax)/(bx-ax);
			} else if (bx < ax) {
				maxArgX = (W-ax)/(bx-ax);
				minArgX = (-W-ax)/(bx-ax);
			} else { // ax = bx
				if ((ax <= W) & (ax >= -W)) {
					maxArgX = 0.0f; // x coordinate is in bounds of screen
					minArgX = 1.0f; // so let it have no influence on lambdA and lambdaB
				} else {
					maxArgX = 2.0f; // line is outside screen
					minArgX = -1.0f; // so force lambdaA > lambdaB
				}
			}
			
			if (by > ay) {
				maxArgY = (-H-ay)/(by-ay);
				minArgY = (H-ay)/(by-ay);
			} else if (by < ay) {
				maxArgY = (H-ay)/(by-ay);
				minArgY = (-H-ay)/(by-ay);
			} else { // ay = by
				if ((ay <= H) & (ay >= -H)) {
					maxArgY = 0.0f; // y coordinate is in bounds of screen
					minArgY = 1.0f; // so let it have no influence on lambdA and lambdaB
				} else {
					maxArgY = 2.0f; // line is outside screen
					minArgY = -1.0f; // so force lambdaA > lambdaB
				}
			}
			
			lambdaA = Math.max(maxArgX, maxArgY);
			lambdaA = Math.max(lambdaA, 0.0f);
			lambdaB = Math.min(minArgX, minArgY);
			lambdaB = Math.min(lambdaB, 1.0f);
			
			// do the following two assignments to align the calculations for
			// axPrime, ayPrime, bxPrime and byPrime for the two cases planar
			// representation is bounded and planar representation wanders off
			// towards infinity
			bx = bx - ax;
			by = by - ay;
			
		} else { // planar representation wanders off towards infinity
		/* we search those lambda such that:
			-W <= ax + lambda * bx
			ax + lambda * bx <= W
			-H <= ay + lambda * by
			ay + lambda * by <= H
			0 <= lambda
		   if, for example, bx > 0 and by > 0:
			lambdaA = max { (-W - ax) / bx, (-H - ay) / by, 0 }
			lambdaB = min { (W - ax) / bx, (H - ay) / by }
		   now if lambdaA <= lambdaB we have some part of the line that
		   intersects with the screen
		*/
			if (bx > 0.0f) {
				maxArgX = (-W-ax)/bx;
				minArgX = (W-ax)/bx;
				minArgXInfinity = false;
			} else if (bx < 0.0f) {
				maxArgX = (W-ax)/bx;
				minArgX = (-W-ax)/bx;
				minArgXInfinity = false;
			} else { // bx = 0.0f
				if ((ax <= W) & (ax >= -W)) {
					maxArgX = 0.0f; // x coordinate is in bounds of screen;
					minArgX = 1.0f; // so let it have no influence on lambdA and lambdaB
									// (minArgX can have any value because we do not use it)
					minArgXInfinity = true;
				} else {
					maxArgX = 2.0f; // line is outside screen
					minArgX = -1.0f; // so force lambdaA > lambdaB
					minArgXInfinity = false;
				}
			}
			
			if (by > 0.0f) {
				maxArgY = (-H-ay)/by;
				minArgY = (H-ay)/by;
				minArgYInfinity = false;
			} else if (by < 0.0f) {
				maxArgY = (H-ay)/by;
				minArgY = (-H-ay)/by;
				minArgYInfinity = false;
			} else { // by = 0.0f
				if ((ay <= H) & (ay >= -H)) {
					maxArgY = 0.0f; // y coordinate is in bounds of screen
					minArgY = 1.0f; // so let it have no influence on lambdA and lambdaB
									// (minArgY can have any value because we do not use it)
					minArgYInfinity = true;
				} else {
					maxArgY = 2.0f; // line is outside screen
					minArgY = -1.0f; // so force lambdaA > lambdaB
					minArgYInfinity =false;
				}
			}
			
			lambdaA = Math.max(maxArgX, maxArgY);
			lambdaA = Math.max(lambdaA, 0.0f);
			
			if (!minArgXInfinity & !minArgYInfinity) {
				lambdaB = Math.min(minArgX, minArgY);
			} else if (!minArgXInfinity & minArgYInfinity) {
				lambdaB = minArgX;
			} else if (minArgXInfinity & !minArgYInfinity) {
				lambdaB = minArgY;
			} else {
				lambdaB = 0.0f;
			}
			
		}
		
		if (lambdaA <= lambdaB) {
			axPrime = ax + lambdaA * bx;
			ayPrime = ay + lambdaA * by;
			bxPrime = ax + lambdaB * bx;
			byPrime = ay + lambdaB * by;
			
			return true;
		} else {
			return false;
		}
		
		
	}
	
	
	// method for finding the screen representation of the branch
	// for the screen plane method
	/*
	public void findScreenRepresentation() {
		int z;
		hasScreenRepresentation = false;
		calculateVisibilitySPM();
		projectOnScreenSPM();
		if (calculatePlanarRepresentationSPM()) {
			if (findIntersectionSPM()) {
				hasScreenRepresentation = true;
				
				z = Math.round(axPrime + Screen.SCREEN_WIDTH/2.0f);
				z = Math.max(z, 0);
				x1 = Math.min(z, Screen.SCREEN_WIDTH-1);
				
				z = Math.round(Screen.SCREEN_HEIGHT/2.0f - ayPrime);
				z = Math.max(z, 0);
				y1 = Math.min(z, Screen.SCREEN_HEIGHT-1);
				
				z = Math.round(bxPrime + Screen.SCREEN_WIDTH/2.0f);
				z = Math.max(z, 0);
				x2 = Math.min(z, Screen.SCREEN_WIDTH-1);
				
				z = Math.round(Screen.SCREEN_HEIGHT/2.0f - byPrime);
				z = Math.max(z, 0);
				y2 = Math.min(z, Screen.SCREEN_HEIGHT-1);
				
			}
		}
	}
	*/
	
	
	// method for finding the screen representation of the branch
	// for the retina plane method
	public void findScreenRepresentation(Screen screen) {
		int z;
		
		// for WARNING messages we determine the distance
		// for the branch end points to the retina plane
		
		Matrix M_lambda = new Matrix(p.subtract(screen.retina),
		 screen.u, screen.v);
		float lambda = Math.abs(M_lambda.determinant());
		if (lambda < 1.0f) {
			System.out.println("WARNING: p distance = " + lambda);
		}
		
		M_lambda = new Matrix(q.subtract(screen.retina),
		 screen.u, screen.v);
		lambda = Math.abs(M_lambda.determinant());
		if (lambda < 1.0f) {
			System.out.println("WARNING: q distance = " + lambda);
		}
		
		
		
		hasScreenRepresentation = false;
		calculateVisibilityRPM(screen);
		projectOnScreenRPM(screen);
		if (calculatePlanarRepresentationRPM(screen)) {
			if (findIntersectionRPM()) {
				hasScreenRepresentation = true;
				
				z = Math.round(axPrime + Screen.SCREEN_WIDTH/2.0f);
				z = Math.max(z, 0);
				x1 = Math.min(z, Screen.SCREEN_WIDTH-1);
				
				z = Math.round(Screen.SCREEN_HEIGHT/2.0f - ayPrime);
				z = Math.max(z, 0);
				y1 = Math.min(z, Screen.SCREEN_HEIGHT-1);
				
				z = Math.round(bxPrime + Screen.SCREEN_WIDTH/2.0f);
				z = Math.max(z, 0);
				x2 = Math.min(z, Screen.SCREEN_WIDTH-1);
				
				z = Math.round(Screen.SCREEN_HEIGHT/2.0f - byPrime);
				z = Math.max(z, 0);
				y2 = Math.min(z, Screen.SCREEN_HEIGHT-1);
				
				determineEnhancementEffect(screen);
			}
		}
	}
	
	// method for calculating the enhancement effect, i.e. width reduction and
	// fade to background
	public void determineEnhancementEffect(Screen screen) {
		// calculate the distance of the visible part of the branch to the
		// screen plane to determine the enhancement effect
		// if delta equals 0 then there is no enhancement effect
		float delta = calculateDistanceToScreen(screen);
		
		// using delta, we determine the width of the branch: the further away,
		// the thinner the branch is
		screenWidth = Tree.MEAN_LENGTH /
		 (Tree.MEAN_LENGTH + Tree.WR_FACTOR * delta) * Tree.PEN_WIDTH;
		
		// using delta, we determine the part of the pen color in the color of
		// the branch: the further away, the smaller this part is and the more
		// the branch fades to the background
		screenPartOfPenColor = Tree.MEAN_LENGTH /
		 (Tree.MEAN_LENGTH + Tree.FTB_FACTOR * delta);
	}
	
	// method for calculating the distance of the visible part of the branch
	// to the screen plane
	// this method shall only be called if hasScreenRepresentation equals true
	public float calculateDistanceToScreen(Screen screen) {
		Vector middle = pPrime.add(qPrime).multiply(0.5f);
		// see the Screen class
		// delta is the distance of the visible part to the screen plane so
		// m - delta n = s^m + mu u + nu v
		// or [n u v] (delta mu nu)^T = m - s^m
		// we use Cramer's rule to calculate delta
		// note that | [n u v] | = 1
		Matrix A_delta = new Matrix(middle.subtract(screen.screenMiddle),
		 screen.u, screen.v);
		float delta = A_delta.determinant();
		// if m is on the same side of the screen plane as the retina plane
		// then delta is negative but now forced to be 0
		// (so it will have no enhancement effect on this branch)
		delta = (float)Math.max(0.0f, delta);
		
		return delta;
	}
	
	/*
	public void setAx(float ax) {
		this.ax = ax;
	}
	
	public void setAy(float ay) {
		this.ay = ay;
	}
	
	public void setBx(float bx) {
		this.bx = bx;
	}
	
	public void setBy(float by) {
		this.by = by;
	}
	
	public void setPPrimeOnPlane(boolean pPrimeOnPlane) {
		this.pPrimeOnPlane = pPrimeOnPlane;
	}
	
	public void setQPrimeOnPlane(boolean qPrimeOnPlane) {
		this.qPrimeOnPlane = qPrimeOnPlane;
	}
	*/
	
}
