/*
A surface is an opaque polygon somewhere in the three dimensional space.
If the surface is planar and convex, it is eligible for rendering.
*/

package tree;

import java.awt.Color;
import java.util.ArrayList;
import display.OpaquePolygon;
import display.Screen;
import geometry.PlanarPolygon;
import geometry.PlanarVector;
import geometry.Polygon;
import geometry.Vector;

public class Surface extends Polygon<Node> {
	
	public static final float EPSILON = 0.001f; // rounding margin
	
	private boolean orientation; // the visible side of the surface has
								 // either counterclockwise orientation (false)
								 // or clockwise orientation (true)
	private Color color;
	private boolean isPlanarSurface; // is this a planar surface?
	private boolean isNormalSurface; // is this a normal surface,
									 // i.e. does it have at least 3 nodes,
									 // are there no 0 rad or pi rad angles,
									 // and is it not self-intersecting?
	private boolean isConvexSurface; // is this a convex surface?
	
	
	public Surface(ArrayList<Node> nodes, boolean orientation, Color color) {
		super(nodes);
		this.orientation = orientation;
		this.color = color;
		
		// we now remove the superfluous nodes
		declutter();
		
		// we determine if the remaining nodes lie in the same plane
		isPlanarSurface = determineIsPlanarSurface();
		
		// we determine if this is a normal surface
		isNormalSurface = determineIsNormalSurface();
		
		// we determine if this is a convex surface
		// (if the surface is not planar or not normal then it is not convex)
		if (isPlanarSurface & isNormalSurface) {
			isConvexSurface = determineIsConvexSurface();
		}
	}
	
	public boolean getOrientation() {
		return orientation;
	}
	
	public Color getColor() {
		return color;
	}
	
	public boolean getIsPlanarSurface() {
		return isPlanarSurface;
	}
	
	public boolean getIsNormalSurface() {
		return isNormalSurface;
	}
	
	public boolean getIsConvexSurface() {
		return isConvexSurface;
	}
	
	private boolean determineIsPlanarSurface() {
		// pre-condition: method declutter() has been called
		if (vertices.size() >= 4) {
		 // only surfaces with at least 4 nodes need to be examined
			
			Vector a = vertices.get(0).subtract(vertices.get(vertices.size()-1));
			 // we know a <> o because of decluttering
			
			// now find b so that a and b are linearly independent
			for (int i=0; i < vertices.size()-1; i++) {
				Vector b = vertices.get(i+1).subtract(vertices.get(i));
				 // we know b <> o because of decluttering
				float cos0 = a.innerProduct(b)/(a.length() * b.length());
				if (Math.abs((Math.abs(cos0) - 1.0f)) >= EPSILON) {
					// we have found b, now check if a, b and c are dependent
					// for all c (we will not check a)
					Vector axb = a.crossProduct(b).normalize();
					 // we know a x b <> o because a and b independent
					for (int j=0; j < vertices.size()-1; j++) {
						Vector c = vertices.get(j+1).subtract(vertices.get(j)).normalize();
						 // we know c <> o because of decluttering
						cos0 = axb.innerProduct(c);
						if (Math.abs(cos0) >= EPSILON) {
							// a, b and c are independent so not planar
							return false;
						}
					}
					// we did not exit the loop with a return false
					// so the surface is planar
					return true;
				}
			}
			// no b could be found so the surface is planar by default
			return true;
		} else {
			// surfaces with less than 4 nodes are planar by default
			return true;
		}
	}
	
	private boolean determineIsNormalSurface() {
		// pre-condition: method declutter() has been called
		
		if (vertices.size() >= 3) {
			Vector a, b; // the edge vectors
			float theta = 0.0f; // the total rotation as we go from edge vector
								// to edge vector
			
			a = vertices.get(0).subtract(vertices.get(vertices.size()-1));
			 // we know a <> o because of decluttering
			for (int i=0; i < vertices.size(); i++) {
				if (i < vertices.size()-1) {
					b = vertices.get(i+1).subtract(vertices.get(i));
				} else {
					b = vertices.get(0).subtract(vertices.get(vertices.size()-1));
				}
				 // we know b <> o because of decluttering
				float cos0 = a.innerProduct(b)/(a.length() * b.length());
				if (Math.abs((Math.abs(cos0) - 1.0f)) >= EPSILON) {
					theta += (float)Math.acos(cos0);
					a = b;
					// angle is neither 0 rad nor pi rad so continue checking
				} else {
					// angle is 0 rad op pi rad so the surface is not normal
					return false;
				}
			}
			// we survived the loop so no 0 rad or pi rad angles
			// but we have to check for theta > 2*pi (self-intersection)
			if (theta <= 1.1f * 2.0f * (float)Math.PI) { // safety margin: 1.1
				return true;
			} else {
				return false;
			}
		} else {
			// a surface with less than 3 nodes is not normal by definition
			return false;
		}
	}
	
	private boolean determineIsConvexSurface() {
		// pre-conditions: method declutter() has been called, and the surface
		// is planar and normal
		if (vertices.size() >= 4) {
			Vector a, b; // the edge vectors
			Vector c, d; // the cross products
			
			a = vertices.get(0).subtract(vertices.get(vertices.size()-1));
			b = vertices.get(1).subtract(vertices.get(0));
			c = a.crossProduct(b);
			 // we know a x b <> o because of decluttering and normal surface
			for (int i=1; i < vertices.size(); i++) {
				a = b;
				if (i < vertices.size()-1) {
					b = vertices.get(i+1).subtract(vertices.get(i));
				} else {
					b = vertices.get(0).subtract(vertices.get(vertices.size()-1));
				}
				d = a.crossProduct(b);
				 // we know a x b <> o because of decluttering and normal surface
				float cos0 = c.innerProduct(d)/(c.length() * d.length());
				if (cos0 < 0.0f) {
					return false;
				}
			}
			// we survived the loop so the surface is convex
			return true;
		} else {
			// surfaces with less than 4 nodes are convex by default
			return true;
		}
	}
	
	// method to determine if the surface is visible for a given screen
	// at this stage in The Tree Project the entire surface has to be on the
	// visible side to be classified as visible
	public boolean isVisible(Screen screen) {
		// check for every node if it is on the visible side
		for (Node node : vertices) {
			if (node.isVisible(screen) == false) {
				return false;
			}
		}
		// we survived the for loop so the surface is visible
		return true;
	}
	
	// method for projecting the surface on the given screen
	// pre-condition: the surface is visible
	public Polygon<Vector> projectOnScreen(Screen screen) {
		ArrayList<Vector> s2Prime = new ArrayList<Vector>();
		for (Node node : vertices) {
			s2Prime.add(node.projectOnScreen(screen));
		}
		return new Polygon<Vector>(s2Prime);
	}
	
	// method to determine if the surface is eligble for rendering
	// given a certain projection on the screen plane
	// pre-condtion: the planar projection has been decluttered
	public boolean hasMatchingOrientation(PlanarPolygon planarProjection) {
		boolean foundCounterClockwise = false;
		boolean foundClockwise = false;
		
		ArrayList<PlanarVector> as = planarProjection.getVertices();
		if (as.size() >= 3) {
			Vector a, b; // the edge vectors
			Vector c; // the cross product
			
			a = as.get(0).subtract(as.get(as.size()-1));
			for (int i=0; i < as.size(); i++) {
				if (i < as.size() -1) {
					b = as.get(i+1).subtract(as.get(i));
				} else {
					b = as.get(0).subtract(as.get(as.size()-1));
				}
				c = a.crossProduct(b);
				// we know a <> o and b <> o because of decluttering
				float sin0 = c.getZ()/(a.length() * b.length());
				if (sin0 > EPSILON) {
					foundCounterClockwise = true;
				} else if (sin0 < -EPSILON) {
					foundClockwise = true;
				}
				a = b;
			}
			
			// now that we have examined all the pairs of adjacent edge vectors
			// we check if the planar projection's orientation matches
			// the surface's orientation
			if (orientation) { // examine clockwise surface
				if (foundClockwise & !foundCounterClockwise) {
					return true;
				} else {
					return false;
				}
			} else { // examine counterclockwise surface
				if (foundCounterClockwise & !foundClockwise) {
					return true;
				} else {
					return false;
				}
			}
		} else {
			// planar projections with less than 3 vertices are not eligible
			// for rendering by default
			return false;
		}
	}
	
	// method for finding the screen representation of the surface
	// the result is returned as an OpaquePolygon object
	// if this object has null value then there is no screen representation
	public OpaquePolygon findScreenRepresentation(Screen screen) {
		
		OpaquePolygon op = null; // the projection of the surface on the screen
								 // in pixel coordinates
		if (isConvexSurface) {
			if (isVisible(screen)) {
				Polygon<Vector> p = projectOnScreen(screen);
				PlanarPolygon pp = p.calculatePlanarRepresentation(screen);
				if (hasMatchingOrientation(pp)) {
					PlanarPolygon cpp = pp.clipToScreen();
					ArrayList<PlanarVector> pvcpp = cpp.getVertices();
					if (pvcpp.size() >= 2) {
						int[] xPoints = new int[pvcpp.size()];
						int[] yPoints = new int[pvcpp.size()];
						int i = 0;
						for (PlanarVector pv : pvcpp) {
							int x = Math.round(pv.getX() + Screen.SCREEN_WIDTH/2.0f);
							x = Math.max(x, 0);
							x = Math.min(x, Screen.SCREEN_WIDTH-1);
							int y = Math.round(Screen.SCREEN_HEIGHT/2.0f - pv.getY());
							y = Math.max(y, 0);
							y = Math.min(y, Screen.SCREEN_HEIGHT-1);
							xPoints[i] = x;
							yPoints[i] = y;
							i++;
						}
						op = new OpaquePolygon(xPoints, yPoints, color);
					}
				}
			}
		}
		return op;
	}
	
}
