/*
A tree is constructed of nodes and branches.
A node is a point somewhere in the three dimensional space.
A branch is a straight line between two nodes.

The main() method of the Tree class is used to start things going.

The static method readTreeFile() is used to read another tree file.

The static method startAnimation() is used to start an animation.

The static branches variable found in class Branches and which is itself of
type Branches is the actual tree, so all Branch objects in that collection are
displayed.

*/


package tree;

import java.awt.Color;
import java.io.File;
import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Locale;
import java.util.Scanner;
import java.util.Set;
import java.util.regex.Pattern;
import gui.AnimationWindow;
import gui.MainWindow;

import static display.Screen.MAIN_SCREEN;

import static tree.ParseState.*;
import static tree.TokenType.*;

public class Tree {
	
	private static MainWindow mainWindow;
	private static AnimationWindow animationWindow;
	
	// constants for general properties
	// some constants are calculated from the collection of branches
	// yet other constants are read from a file
	// and finally there are constants which are calculated from the constants
	// read from the file
	// for a constant which can be read from a file, an attempt is made to read
	// a proper value from the file but if this fails then a default value will
	// be used
	
	// constants calculated from collection of branches
	
	public static int NUMBER_OF_BRANCHES; // the number of branches
	
	public static float MEAN_LENGTH; // the arithmetic mean of all branch lenghts
	
	// constants read from file (and their default values)
	
	private static int BG_COLOR_DEFAULT = 0x00003F;
	public static int BG_COLOR = BG_COLOR_DEFAULT; // the background color
	
	private static int PEN_COLOR_DEFAULT = 0xFFFF00;
	public static int PEN_COLOR = PEN_COLOR_DEFAULT; // the pen color for the wireframe lines
	
	private static float PEN_WIDTH_DEFAULT = 2.0f;
	public static float PEN_WIDTH = PEN_WIDTH_DEFAULT; // the pen thickness for the wireframe lines
	
	private static float FTB_FACTOR_DEFAULT = 0.0f;
	public static float FTB_FACTOR = FTB_FACTOR_DEFAULT; // the fade to background factor
	
	private static float WR_FACTOR_DEFAULT = 0.0f;
	public static float WR_FACTOR = WR_FACTOR_DEFAULT; // the width reduction factor
	
	// constants calculated from constants read from file
	
	public static int BG_RED;    // red part of background color
	public static int BG_GREEN;  // green part of background color
	public static int BG_BLUE;   // blue part of background color
	public static int PEN_RED;   // red part of pen color
	public static int PEN_GREEN; // green part of pen color
	public static int PEN_BLUE;  // blue part of pen color
	
	
	public static void main(String[] args) {
		
		// the main method creates a default tree, this tree is a pyramid
		
		Nodes nodes = Nodes.getInstance();
		Branches branches = Branches.getInstance();
		
		// pyramid:
		nodes.put("pn1", new Node(-250.0f, 0.0f, 1000.0f));
		nodes.put("pn2", new Node(250.0f, 0.0f, 1000.0f));
		nodes.put("pn3", new Node(250.0f, 0.0f, 1500.0f));
		nodes.put("pn4", new Node(-250.0f, 0.0f, 1500.0f));
		nodes.put("pn5", new Node(0.0f, 500.0f, 1250.0f));
		
		branches.put("pb1", new Branch(nodes.get("pn1"), nodes.get("pn2")));
		branches.put("pb2", new Branch(nodes.get("pn2"), nodes.get("pn3")));
		branches.put("pb3", new Branch(nodes.get("pn3"), nodes.get("pn4")));
		branches.put("pb4", new Branch(nodes.get("pn4"), nodes.get("pn1")));
		branches.put("pb5", new Branch(nodes.get("pn1"), nodes.get("pn5")));
		branches.put("pb6", new Branch(nodes.get("pn2"), nodes.get("pn5")));
		branches.put("pb7", new Branch(nodes.get("pn3"), nodes.get("pn5")));
		branches.put("pb8", new Branch(nodes.get("pn4"), nodes.get("pn5")));
		
		
		// before the GUI can be started, several tasks have to be done
		
		// task 1: calculate constants from collection of branches
		calculateConstantsFromBranches();
		
		// task 2: read constants from file
		readConstantsFromFile();
		
		// task 3: calculate constants from constants read from file
		calculateConstantsFromFileConstants();
		
		// task 4: find the screen representation
		MAIN_SCREEN.findRepresentation();
		
		// when all tasks are done: start the GUI
		mainWindow = new MainWindow();
		
		// run a short animation
		mainWindow.doAnimation();
		
	}
	
	// static method readTreeFile() is used to read a new tree from a so-called
	// tree file, it implements the use case "Parse tree file", which is
	// modeled by a state machine
	public static int readTreeFile(File treeFile) {
		// return the number of parse errors
		int numberOfErrors = 0;
		
		try {
			Scanner s = new Scanner(treeFile);
			
			// most of the following method variables need to be initialized
			// to prevent compiler errors ("variable might not have been
			// initialized")
			String token = null;
			TokenType tokenType;
			String nodeName = null, branchName = null;
			float nodeX = 0.0f, nodeY = 0.0f, nodeZ;
			Node branchP = null, branchQ;
			Color color = new Color(0);
			boolean orientation = false;
			ArrayList<Node> surfaceNodes = new ArrayList<Node>();
			
			// get new build instances for nodes, branches and surfaces
			Nodes nodes = Nodes.createBuildInstance();
			Branches branches = Branches.createBuildInstance();
			Surfaces surfaces = Surfaces.createBuildInstance();
			
			// start the state machine
			ParseState parseState = IGNORE_INPUT;
			do {
				// check if the next token is a comment...
				// (a comment starts with a '!')
				while (s.hasNext(Pattern.compile("!.*"))) {
					// because a comment was found skip the remainder of the line
					// but note that the comment itself may be on another line
					// (in which case the above condition will be true once more)
					s.nextLine();
				}
				
				if (s.hasNext()) {
					token = s.next();
					if (token.equalsIgnoreCase("[NODES]")) {
						tokenType = NODES;
					} else if (token.equalsIgnoreCase("[BRANCHES]")) {
						tokenType = BRANCHES;
					} else if (token.equalsIgnoreCase("[SURFACE]")) {
						tokenType = SURFACE;
					} else {
						tokenType = REGULAR;
					}
				} else {
					tokenType = EOF;
				}
				
				switch (parseState) {
				case IGNORE_INPUT:
					switch (tokenType) {
					case REGULAR:
						numberOfErrors++;
						System.err.println("INFO: invalid token");
						// stay in IGNORE_INPUT
						break;
					case NODES:
						parseState = READ_NODE_NAME;
						break;
					case BRANCHES:
						parseState = READ_BRANCH_NAME;
						break;
					case SURFACE:
						parseState = READ_COLOR;
						break;
					case EOF:
						break;
					}
					break;
				case READ_NODE_NAME:
					switch (tokenType) {
					case REGULAR:
						nodeName = token;
						parseState = READ_NODE_X;
						break;
					case NODES:
						// stay in READ_NODE_NAME
						break;
					case BRANCHES:
						parseState = READ_BRANCH_NAME;
						break;
					case SURFACE:
						parseState = READ_COLOR;
						break;
					case EOF:
						break;
					}
					break;
				case READ_NODE_X:
					switch (tokenType) {
					case REGULAR:
						try {
							nodeX = Float.parseFloat(token);
							parseState = READ_NODE_Y;
						} catch(NumberFormatException nfe) {
							numberOfErrors++;
							System.err.println("INFO: invalid (not a float) x coordinate for node "
							 + nodeName);
							nodeName = token;
							// stay in READ_NODE_X
						}
						break;
					case NODES:
						numberOfErrors++;
						System.err.println("INFO: missing x, y and z coordinates for node "
						 + nodeName);
						parseState = READ_NODE_NAME;
						break;
					case BRANCHES:
						numberOfErrors++;
						System.err.println("INFO: missing x, y and z coordinates for node "
						 + nodeName);
						parseState = READ_BRANCH_NAME;
						break;
					case SURFACE:
						numberOfErrors++;
						System.err.println("INFO: missing x, y and z coordinates for node "
						 + nodeName);
						parseState = READ_COLOR;
						break;
					case EOF:
						numberOfErrors++;
						System.err.println("INFO: missing x, y and z coordinates for node "
						 + nodeName);
						break;
					}
					break;
				case READ_NODE_Y:
					switch (tokenType) {
					case REGULAR:
						try {
							nodeY = Float.parseFloat(token);
							parseState = READ_NODE_Z;
						} catch(NumberFormatException nfe) {
							numberOfErrors++;
							System.err.println("INFO: invalid (not a float) y coordinate for node "
							 + nodeName);
							nodeName = token;
							parseState = READ_NODE_X;
						}
						break;
					case NODES:
						numberOfErrors++;
						System.err.println("INFO: missing y and z coordinates for node "
						 + nodeName);
						parseState = READ_NODE_NAME;
						break;
					case BRANCHES:
						numberOfErrors++;
						System.err.println("INFO: missing y and z coordinates for node "
						 + nodeName);
						parseState = READ_BRANCH_NAME;
						break;
					case SURFACE:
						numberOfErrors++;
						System.err.println("INFO: missing y and z coordinates for node "
						 + nodeName);
						parseState = READ_COLOR;
						break;
					case EOF:
						numberOfErrors++;
						System.err.println("INFO: missing y and z coordinates for node "
						 + nodeName);
						break;
					}
					break;
				case READ_NODE_Z:
					switch (tokenType) {
					case REGULAR:
						try {
							nodeZ = Float.parseFloat(token);
							Node node = new Node(nodeX, nodeY, nodeZ);
							nodes.put(nodeName, node);
							parseState = READ_NODE_NAME;
						} catch (NumberFormatException nfe) {
							numberOfErrors++;
							System.err.println("INFO: invalid (not a float) z coordinate for node "
							 + nodeName);
							nodeName = token;
							parseState = READ_NODE_X;
						}
						break;
					case NODES:
						numberOfErrors++;
						System.err.println("INFO: missing z coordinate for node "
						 + nodeName);
						parseState = READ_NODE_NAME;
						break;
					case BRANCHES:
						numberOfErrors++;
						System.err.println("INFO: missing z coordinate for node "
						 + nodeName);
						parseState = READ_BRANCH_NAME;
						break;
					case SURFACE:
						numberOfErrors++;
						System.err.println("INFO: missing z coordinate for node "
						 + nodeName);
						parseState = READ_COLOR;
						break;
					case EOF:
						numberOfErrors++;
						System.err.println("INFO: missing z coordinate for node "
						 + nodeName);
						break;
					}
					break;
				case READ_BRANCH_NAME:
					switch (tokenType) {
					case REGULAR:
						branchName = token;
						parseState = READ_BRANCH_P;
						break;
					case NODES:
						parseState = READ_NODE_NAME;
						break;
					case BRANCHES:
						// stay in READ_BRANCH_NAME
						break;
					case SURFACE:
						parseState = READ_COLOR;
					case EOF:
						break;
					}
					break;
				case READ_BRANCH_P:
					switch (tokenType) {
					case REGULAR:
						if (nodes.containsKey(token)) {
							branchP = nodes.get(token);
							parseState = READ_BRANCH_Q;
						} else {
							numberOfErrors++;
							System.err.println("INFO: invalid (non-existent) p node for branch "
							 + branchName);
							branchName = token;
							// stay in READ_BRANCH_P
						}
						break;
					case NODES:
						numberOfErrors++;
						System.err.println("INFO: missing p and q nodes for branch "
						 + branchName);
						parseState = READ_NODE_NAME;
						break;
					case BRANCHES:
						numberOfErrors++;
						System.err.println("INFO: missing p and q nodes for branch "
						 + branchName);
						parseState = READ_BRANCH_NAME;
						break;
					case SURFACE:
						numberOfErrors++;
						System.err.println("INFO: missing p and q nodes for branch "
						 + branchName);
						parseState = READ_COLOR;
						break;
					case EOF:
						numberOfErrors++;
						System.err.println("INFO: missing p and q nodes for branch "
						 + branchName);
						break;
					}
					break;
				case READ_BRANCH_Q:
					switch (tokenType) {
					case REGULAR:
						if (nodes.containsKey(token)) {
							branchQ = nodes.get(token);
							Branch branch = new Branch(branchP, branchQ);
							branches.put(branchName, branch);
							parseState = READ_BRANCH_NAME;
						} else {
							numberOfErrors++;
							System.err.println("INFO: invalid (non-existent) q node for branch "
							 + branchName);
							branchName = token;
							parseState = READ_BRANCH_P;
						}
						break;
					case NODES:
						numberOfErrors++;
						System.err.println("INFO: missing q node for branch "
						 + branchName);
						parseState = READ_NODE_NAME;
						break;
					case BRANCHES:
						numberOfErrors++;
						System.err.println("INFO: missing q node for branch "
						 + branchName);
						parseState = READ_BRANCH_NAME;
						break;
					case SURFACE:
						numberOfErrors++;
						System.err.println("INFO: missing q node for branch "
						 + branchName);
						parseState = READ_COLOR;
						break;
					case EOF:
						numberOfErrors++;
						System.err.println("INFO: missing q node for branch "
						 + branchName);
						break;
					}
					break;
				case READ_COLOR:
					switch (tokenType) {
					case REGULAR:
						try {
							int colorValue = Integer.parseInt(token);
							if (colorValue < 0 | colorValue > 16777215) {
								throw new NumberFormatException();
							}
							color = new Color(colorValue);
							parseState = READ_ORIENTATION;
						} catch(NumberFormatException nfe) {
							numberOfErrors++;
							System.err.println("INFO: invalid (not an integer in the range from 0 to 16777215) color for surface");
							parseState = IGNORE_INPUT;
						}
						break;
					case NODES:
						numberOfErrors++;
						System.err.println("INFO: missing color and orientation for surface");
						parseState = READ_NODE_NAME;
						break;
					case BRANCHES:
						numberOfErrors++;
						System.err.println("INFO: missing color and orientation for surface");
						parseState = READ_BRANCH_NAME;
						break;
					case SURFACE:
						numberOfErrors++;
						System.err.println("INFO: missing color and orientation for surface");
						// stay in READ_COLOR
						break;
					case EOF:
						numberOfErrors++;
						System.err.println("INFO: missing color and orientation for surface");
						break;
					}
					break;
				case READ_ORIENTATION:
					switch (tokenType) {
					case REGULAR:
						if (token.equalsIgnoreCase("NEGATIVE")) {
							orientation = false;
							surfaceNodes = new ArrayList<Node>();
							// make sure to start with a fresh list
							parseState = READ_SURFACE_P;
						} else if (token.equalsIgnoreCase("POSITIVE")) {
							orientation = true;
							surfaceNodes = new ArrayList<Node>();
							// make sure to start with a fresh list
							parseState = READ_SURFACE_P;
						} else {
							numberOfErrors++;
							System.err.println("INFO: invalid (neither NEGATIVE nor POSITIVE) orientation for surface");
							parseState = IGNORE_INPUT;
						}
						break;
					case NODES:
						numberOfErrors++;
						System.err.println("INFO: missing orientation for surface");
						parseState = READ_NODE_NAME;
						break;
					case BRANCHES:
						numberOfErrors++;
						System.err.println("INFO: missing orientation for surface");
						parseState = READ_BRANCH_NAME;
						break;
					case SURFACE:
						numberOfErrors++;
						System.err.println("INFO: missing orientation for surface");
						parseState = READ_COLOR;
						break;
					case EOF:
						numberOfErrors++;
						System.err.println("INFO: missing orientation for surface");
						break;
					}
					break;
				case READ_SURFACE_P:
					switch (tokenType) {
					case REGULAR:
						if (nodes.containsKey(token)) {
							surfaceNodes.add(nodes.get(token));
							// stay in READ_SURFACE_P;
						} else {
							numberOfErrors++;
							System.err.println("INFO: invalid (non-existent) node for surface");
							parseState = IGNORE_INPUT;
						}
						break;
					case NODES:
						surfaces.add(new Surface(surfaceNodes, orientation, color));
						parseState = READ_NODE_NAME;
						break;
					case BRANCHES:
						surfaces.add(new Surface(surfaceNodes, orientation, color));
						parseState = READ_BRANCH_NAME;
						break;
					case SURFACE:
						surfaces.add(new Surface(surfaceNodes, orientation, color));
						parseState = READ_COLOR;
						break;
					case EOF:
						surfaces.add(new Surface(surfaceNodes, orientation, color));
						break;
					}
					break;
				}
			} while (tokenType != EOF);
			s.close();
			System.err.println("INFO: number of parse errors is " +
			 numberOfErrors);
		} catch (FileNotFoundException e) {
			System.err.println("WARNING: unable to open file " +
			 treeFile.getName());
			return 0; // unable to open file so make sure that no parse error
					  // warning will be shown (this exception may be dealt
					  // with in a better way in a next version)
		}
		
		// before the new tree can be displayed, several tasks have to be done
		
		// task 1: replace the singleton classes with the build instances
		Nodes.commitBuildInstance();
		Branches.commitBuildInstance();
		Surfaces.commitBuildInstance();
		
		// task 2: recalculate constants from collection of branches
		calculateConstantsFromBranches();
		
		// task 3: reset the viewpoint to the initial viewpoint
		MAIN_SCREEN.resetViewpoint();
		
		// task 4: find the screen representation
		MAIN_SCREEN.findRepresentation();
		
		// when all tasks are done: repaint
		mainWindow.getTreeCanvas().repaint();
		
		return numberOfErrors;
		
	}
	
	
	private static void calculateConstantsFromBranches() {
		Branches branches = Branches.getInstance();
		NUMBER_OF_BRANCHES = branches.size();
		
		if (NUMBER_OF_BRANCHES != 0) {
			float sum = 0.0f;
			for (Branch branch : branches.values()) {
				sum = sum + branch.getNode1().subtract(branch.getNode2()).length();
			}
			MEAN_LENGTH = sum / NUMBER_OF_BRANCHES;
		} else {
			MEAN_LENGTH = 0.0f;
		}
	}
	
	// method for reading constants from a file
	// the path to the file should be
	// config/properties.txt from the working directory
	// (../config/properties.txt from this directory)
	private static void readConstantsFromFile() {
		
		boolean bgColorPropertyExists = false;
		boolean penColorPropertyExists = false;
		boolean penWidthPropertyExists = false;
		boolean ftbFactorPropertyExists = false;
		boolean wrFactorPropertyExists = false;
		
		File propertiesFile = new File("config/properties.txt");
		try {
			Scanner s = new Scanner(propertiesFile);
			
			// set the Scanner object's locale to English
			// this is done because for example in my case the default locale
			// is Dutch which causes wrong interpretation of floating points
			// (for example you write 0.1 in English but 0,1 in Dutch)
			s.useLocale(Locale.ENGLISH);
			
			while (s.hasNext()) {
				
				// check if the next token is a comment...
				// (a comment starts with a '!')
				if (s.hasNext(Pattern.compile("!.*"))) {
					// because a comment was found skip the remainder of the line
					// but note that the comment itself may be on another line
					// (in which case the above condition will be true once more)
					s.nextLine();
					
				// ... or a candidate for a property
				} else {
					String property = s.next();
					if (property.equalsIgnoreCase("BG_COLOR")) {
						if (s.hasNextInt()) {
							BG_COLOR = s.nextInt();
							bgColorPropertyExists = true;
						}
					} else if (property.equalsIgnoreCase("PEN_COLOR")) {
						if (s.hasNextInt()) {
							PEN_COLOR = s.nextInt();
							penColorPropertyExists = true;
						}
					} else if (property.equalsIgnoreCase("PEN_WIDTH")) {
						if (s.hasNextFloat()) {
							PEN_WIDTH = s.nextFloat();
							penWidthPropertyExists = true;
						}
					} else if (property.equalsIgnoreCase("FTB_FACTOR")) {
						if (s.hasNextFloat()) {
							FTB_FACTOR = s.nextFloat();
							ftbFactorPropertyExists = true;
						}
					} else if (property.equalsIgnoreCase("WR_FACTOR")) {
						if (s.hasNextFloat()) {
							WR_FACTOR = s.nextFloat();
							wrFactorPropertyExists = true;
						}
					}
				}
			}
			
			s.close();
			
		} catch (FileNotFoundException e) {
			System.err.println("WARNING: unable to find or open file properties.txt");
		}
		
		if (!bgColorPropertyExists) {
			System.err.println("INFO: no proper value read for BG_COLOR, reverting to default value");
		}
		
		if (!penColorPropertyExists) {
			System.err.println("INFO: no proper value read for PEN_COLOR, reverting to default value");
		}
		
		if (!penWidthPropertyExists) {
			System.err.println("INFO: no proper value read for PEN_WIDTH, reverting to default value");
		}
		
		if (!ftbFactorPropertyExists) {
			System.err.println("INFO: no proper value read for FTB_FACTOR, reverting to default value");
		}
		
		if (!wrFactorPropertyExists) {
			System.err.println("INFO: no proper value read for WR_FACTOR, reverting to default value");
		}
	}
	
	private static void calculateConstantsFromFileConstants() {
		
		BG_RED    = (BG_COLOR  & 0xFF0000) >> 16;
		BG_GREEN  = (BG_COLOR  & 0x00FF00) >> 8;
		BG_BLUE   = (BG_COLOR  & 0x0000FF);
		PEN_RED   = (PEN_COLOR & 0xFF0000) >> 16;
		PEN_GREEN = (PEN_COLOR & 0x00FF00) >> 8;
		PEN_BLUE  = (PEN_COLOR & 0x0000FF);
		
	}
	
	
	// static method startAnimation() is called when the user starts an
	// animation
	public static void startAnimation() {
		if (!AnimationWindow.topLevelWindowExists) {
			// start the GUI if top level window does not exist
			animationWindow = new AnimationWindow();
		}
		
		// run the animation
		animationWindow.doAnimation();
	}
	
	
	private static void printTree() {
		
		Branches branches = Branches.getInstance();
		
		System.out.println("Branches in this tree are:");
		Set<String> branchNames = branches.keySet();
		for (String branchName : branchNames) {
			System.out.println("branch: " + branchName);
		}
		System.out.println();
		
		System.out.println("Nodes per branch are:");
		for (String branchName : branchNames) {
			System.out.println("branch " + branchName + " contains:");
			Branch branch = branches.get(branchName);
			System.out.println(" node1: (" + branch.getNode1().getX() +
										", " + branch.getNode1().getY() +
										", " + branch.getNode1().getZ() +
										")");
			System.out.println(" node2: (" + branch.getNode2().getX() +
										", " + branch.getNode2().getY() +
										", " + branch.getNode2().getZ() +
										")");
		}
		
	}
	
}
