/*
A trail is constructed of markers and trail segments.
A marker is a point somewhere in the three dimensional space. It also has a
(negatively oriented) orthonormal base.
A trail segment is a path from a start marker to an end marker.

A marker is a visual aid. It is a node but it is not intended to be used as a
tree construct, instead it is used for setting up an animation.
A marker has a fixed size and a fixed color. The display size is adjusted, but
the display color is not.
*/

package visual_aid;

import display.Screen;
import geometry.Matrix;
import geometry.Vector;
import geometry.PlanarVector;
import tree.Node;

import static display.Screen.MAIN_SCREEN;

public class Marker extends Node {
	
	private final Vector u, v, normal; // once a Marker object has been
									   // created, it cannot be changed
	
	private Vector m2Prime;
	// the projection of the marker on the screen plane
	//
	// if the marker is on the visible side of the retina plane then projection
	// can be done
	// if the marker is not on the visible side of the retina plane then
	// m2Prime will get null value
	
	private PlanarVector a;
	// 2D representation of the projection of the marker in terms of u and v
	//
	// if m2Prime is unequal to null, then
	// m2Prime - screenMiddle = a_x * u + a_y * v
	
	private float screenSize;
	// the size of the marker on the screen taking into account the enhancement
	// effect
	
	private boolean hasScreenRepresentation;
	// indicates if the marker is (partly) visible or not
	
	private float x1, y1;
	// pixel coordinates of the marker's center
	// (the marker will be drawn as a disc)
	
	// the constructor without arguments uses the vectors s^m, u, v and n
	public Marker() {
		super(MAIN_SCREEN.screenMiddle);
		u = MAIN_SCREEN.u;
		v = MAIN_SCREEN.v;
		normal = MAIN_SCREEN.normal;
	}
	
	// the constructor with arguments uses them as they are
	public Marker(Vector p, Vector u, Vector v, Vector normal) {
		super(p);
		this.u = u;
		this.v = v;
		this.normal = normal;
	}
	
	public Vector getU() {
		return u;
	}
	
	public Vector getV() {
		return v;
	}
	
	public Vector getNormal() {
		return normal;
	}
	
	public Vector getM2Prime() {
		return m2Prime;
	}
	
	public PlanarVector getA() {
		return a;
	}
	
	public float getScreenSize() {
		return screenSize;
	}
	
	public boolean getHasScreenRepresentation() {
		return hasScreenRepresentation;
	}
	
	public float getX1() {
		return x1;
	}
	
	public float getY1() {
		return y1;
	}
	
	
	
	
	// method for calculating if there is an intersection of the 2D planar
	// representation with the screen, this involves calculating screenSize
	// returns true if the representation has an intersection with the screen
	// returns false otherwise
	// pre-condition: the marker is visible
	private boolean findIntersection(Screen screen) {
		// calculate the distance of the marker to the screen plane to
		// determine the enhancement effect
		// if delta equals 0 then there is no enhancement effect
		float delta = calculateDistanceToScreen(screen);
		
		screenSize = Screen.RETINA_SCREEN_DISTANCE /
		 (Screen.RETINA_SCREEN_DISTANCE + delta) * Markers.MARKER_SIZE;
		if ((a.getX() >= -Screen.SCREEN_WIDTH/2.0f - screenSize) &&
			(a.getX() <= Screen.SCREEN_WIDTH/2.0f + screenSize) &&
			(a.getY() >= -Screen.SCREEN_HEIGHT/2.0f - screenSize) &&
			(a.getY() <= Screen.SCREEN_HEIGHT/2.0f + screenSize)) {
			return true;
		} else {
			return false;
		}
	}
	
	// method for calculating the distance of the marker to the screen plane
	// pre-condition: the marker is visible
	private float calculateDistanceToScreen(Screen screen) {
		// see the Screen class
		// delta is the distance of the marker m to the screen plane so
		// m - delta n = s^m + mu u + nu v
		// or [n u v] (delta mu nu)^T = m - s^m
		// we use Cramer's rule to calculate delta
		// note that | [n u v] | = 1
		Matrix A_delta = new Matrix(this.subtract(screen.screenMiddle),
		 screen.u, screen.v);
		float delta = A_delta.determinant();
		// if the marker is on the same side of the screen plane as the retina
		// plane then delta is negative but now forced to be 0
		// (so it will have no enhancement effect on this marker)
		delta = (float)Math.max(0.0f, delta);
		
		return delta;
	}
	
	// method for finding the screen representation of the marker
	public void findScreenRepresentation(Screen screen) {
		hasScreenRepresentation = false;
		if (isVisible(screen) == true) {
			m2Prime = projectOnScreen(screen);
			a = m2Prime.calculatePlanarRepresentation(screen);
			if (findIntersection(screen)) {
				hasScreenRepresentation = true;
				x1 = Screen.SCREEN_WIDTH/2.0f + a.getX();
				y1 = Screen.SCREEN_HEIGHT/2.0f - a.getY();
			}
		} else {
			m2Prime = null;
		}
	}
	
}
