/*
A Markers object is an ArrayList of Marker objects.
This class has several static variables which are properties of all markers:
- The flag showMarkersInManualMode determines whether markers will be shown in
manual mode.
- The flag showMarkersInAnimationMode determines whether markers will be shown
in animation mode.
- The Marker selectedMarker stores the reference to the selected marker object
(if no marker is selected then it has null value).
*/

package visual_aid;

import geometry.Vector;
import gui.MainWindow;

import java.awt.*;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Scanner;
import java.util.regex.Pattern;

import static display.Screen.MAIN_SCREEN;

import static visual_aid.ParseState.*;
import static visual_aid.TokenType.*;

public class Markers extends ArrayList<Marker> {
	
	// flags:
	
	public static final boolean SHOW_MARKERS_IN_MANUAL_MODE = true;
	public static boolean showMarkersInManualMode = SHOW_MARKERS_IN_MANUAL_MODE;
	// flags whether markers shall be shown during manual screen control
	
	public static final boolean SHOW_MARKERS_IN_ANIMATION_MODE = false;
	public static boolean showMarkersInAnimationMode = SHOW_MARKERS_IN_ANIMATION_MODE;
	// flags whether markers shall be shown during animated screen control
	
	public static Marker selectedMarker = null;
	// selected marker
	
	// constants for general properties
	// constants are read from a file
	// for a constant which can be read from a file, an attempt is made to read
	// a proper value from the file but if this fails then a default value will
	// be used
	
	// constants read from file (and their default values)
	
	private static int MARKER_COLOR_DEFAULT = 0xFF0000; // the pen color for the markers
	public static int MARKER_COLOR = MARKER_COLOR_DEFAULT;
	
	private static int SELECTED_MARKER_COLOR_DEFAULT = 0x00FF00; // the pen color for the selected marker
	public static int SELECTED_MARKER_COLOR = SELECTED_MARKER_COLOR_DEFAULT;
	
	private static float FIRST_KEYFRAME_DEFAULT = 0.9f; // the mark up to which the pen color shall be opaque
	public static float FIRST_KEYFRAME = FIRST_KEYFRAME_DEFAULT;
	
	private static float MARKER_SIZE_DEFAULT = 15 / FIRST_KEYFRAME; // the size for the markers
	public static float MARKER_SIZE = MARKER_SIZE_DEFAULT;
	// note that we divide by FIRST_KEYFRAME to increase the size, the opaque
	// part will then be equal to the intended disc size
	
	// Markers is implemented as a singleton class
	private static Markers markers = new Markers();
	// but it has another Markers object used to build a new animation
	private static Markers buildInstance;
	
	// A private constructor prevents any other class from instantiating
	private Markers() {}
	
	// A method to get the reference to the singleton class
	public static Markers getInstance() {
		return markers;
	}
	
	// static method loadAnimationFile() is used to load an animation from a
	// so-called animation file, it implements the use case "Parse animation
	// file", which is modeled by a state machine
	public static int loadAnimationFile(File animationFile, MainWindow mainWindow) {
		// return the number of parse errors
		int numberOfErrors = 0;
		// start the state machine
		ParseState parseState = READ_MARKER;
		
		try {
			Scanner s = new Scanner(animationFile);
			
			// some of the following method variables need to be initialized
			// only to prevent compiler errors ("variable might not have been
			// initialized")
			String token = null;
			TokenType tokenType;
			float p_x=0.0f, p_y=0.0f, p_z=0.0f, u_x=0.0f, u_y=0.0f, u_z=0.0f,
			 v_x=0.0f, v_y=0.0f, v_z=0.0f, n_x=0.0f, n_y=0.0f, n_z;
			
			// get new build instance for animation
			buildInstance = new Markers();
			
			do {
				// check if the next token is a comment...
				// (a comment starts with a '!')
				while (s.hasNext(Pattern.compile("!.*"))) {
					// because a comment was found skip the remainder of the line
					// but note that the comment itself may be on another line
					// (in which case the above condition will be true once more)
					s.nextLine();
				}
				
				if (s.hasNext()) {
					token = s.next();
					tokenType = REGULAR;
				} else {
					tokenType = EOF;
				}
				
				switch (tokenType) {
				case REGULAR:
					switch (parseState) {
					case READ_MARKER:
						if (token.equalsIgnoreCase("[MARKER]")) {
							parseState = READ_X;
						} else {
							numberOfErrors++;
							System.err.println("INFO: [MARKER] expected");
							// stay in READ_MARKER
						}
						break;
					case READ_X:
						if (token.equalsIgnoreCase("[MARKER]")) {
							numberOfErrors++;
							System.err.println("INFO: data from p_x onwards missing");
							// stay in READ_X
						} else {
							try {
								p_x = Float.parseFloat(token);
								parseState = READ_Y;
							} catch (NumberFormatException nfe) {
								numberOfErrors++;
								System.err.println("INFO: invalid (not a float) p_x coordinate");
								parseState = READ_MARKER;
							}
						}
						break;
					case READ_Y:
						if (token.equalsIgnoreCase("[MARKER]")) {
							numberOfErrors++;
							System.err.println("INFO: data from p_y onwards missing");
							parseState = READ_X;
						} else {
							try {
								p_y = Float.parseFloat(token);
								parseState = READ_Z;
							} catch (NumberFormatException nfe) {
								numberOfErrors++;
								System.err.println("INFO: invalid (not a float) p_y coordinate");
								parseState = READ_MARKER;
							}
						}
						break;
					case READ_Z:
						if (token.equalsIgnoreCase("[MARKER]")) {
							numberOfErrors++;
							System.err.println("INFO: data from p_z onwards missing");
							parseState = READ_X;
						} else {
							try {
								p_z = Float.parseFloat(token);
								parseState = READ_U;
							} catch (NumberFormatException nfe) {
								numberOfErrors++;
								System.err.println("INFO: invalid (not a float) p_z coordinate");
								parseState = READ_MARKER;
							}
						}
						break;
					case READ_U:
						if (token.equalsIgnoreCase("[MARKER]")) {
							numberOfErrors++;
							System.err.println("INFO: data from [U} onwards missing");
							parseState = READ_X;
						} else if (token.equalsIgnoreCase("[U]")) {
							parseState = READ_U_X;
						} else {
							numberOfErrors++;
							System.err.println("INFO: [U] expected");
							parseState = READ_MARKER;
						}
						break;
					case READ_U_X:
						if (token.equalsIgnoreCase("[MARKER]")) {
							numberOfErrors++;
							System.err.println("INFO: data from u_x onwards missing");
							parseState = READ_X;
						} else {
							try {
								u_x = Float.parseFloat(token);
								parseState = READ_U_Y;
							} catch (NumberFormatException nfe) {
								numberOfErrors++;
								System.err.println("INFO: invalid (not a float) u_x coordinate");
								parseState = READ_MARKER;
							}
						}
						break;
					case READ_U_Y:
						if (token.equalsIgnoreCase("[MARKER]")) {
							numberOfErrors++;
							System.err.println("INFO: data from u_y onwards missing");
							parseState = READ_X;
						} else {
							try {
								u_y = Float.parseFloat(token);
								parseState = READ_U_Z;
							} catch (NumberFormatException nfe) {
								numberOfErrors++;
								System.err.println("INFO: invalid (not a float) u_y coordinate");
								parseState = READ_MARKER;
							}
						}
						break;
					case READ_U_Z:
						if (token.equalsIgnoreCase("[MARKER]")) {
							numberOfErrors++;
							System.err.println("INFO: data from u_z onwards missing");
							parseState = READ_X;
						} else {
							try {
								u_z = Float.parseFloat(token);
								parseState = READ_V;
							} catch (NumberFormatException nfe) {
								numberOfErrors++;
								System.err.println("INFO: invalid (not a float) u_z coordinate");
								parseState = READ_MARKER;
							}
						}
						break;
					case READ_V:
						if (token.equalsIgnoreCase("[MARKER]")) {
							numberOfErrors++;
							System.err.println("INFO: data from [V} onwards missing");
							parseState = READ_X;
						} else if (token.equalsIgnoreCase("[V]")) {
							parseState = READ_V_X;
						} else {
							numberOfErrors++;
							System.err.println("INFO: [V] expected");
							parseState = READ_MARKER;
						}
						break;
					case READ_V_X:
						if (token.equalsIgnoreCase("[MARKER]")) {
							numberOfErrors++;
							System.err.println("INFO: data from v_x onwards missing");
							parseState = READ_X;
						} else {
							try {
								v_x = Float.parseFloat(token);
								parseState = READ_V_Y;
							} catch (NumberFormatException nfe) {
								numberOfErrors++;
								System.err.println("INFO: invalid (not a float) v_x coordinate");
								parseState = READ_MARKER;
							}
						}
						break;
					case READ_V_Y:
						if (token.equalsIgnoreCase("[MARKER]")) {
							numberOfErrors++;
							System.err.println("INFO: data from v_y onwards missing");
							parseState = READ_X;
						} else {
							try {
								v_y = Float.parseFloat(token);
								parseState = READ_V_Z;
							} catch (NumberFormatException nfe) {
								numberOfErrors++;
								System.err.println("INFO: invalid (not a float) v_y coordinate");
								parseState = READ_MARKER;
							}
						}
						break;
					case READ_V_Z:
						if (token.equalsIgnoreCase("[MARKER]")) {
							numberOfErrors++;
							System.err.println("INFO: data from v_z onwards missing");
							parseState = READ_X;
						} else {
							try {
								v_z = Float.parseFloat(token);
								parseState = READ_N;
							} catch (NumberFormatException nfe) {
								numberOfErrors++;
								System.err.println("INFO: invalid (not a float) v_z coordinate");
								parseState = READ_MARKER;
							}
						}
						break;
					case READ_N:
						if (token.equalsIgnoreCase("[MARKER]")) {
							numberOfErrors++;
							System.err.println("INFO: data from [N} onwards missing");
							parseState = READ_X;
						} else if (token.equalsIgnoreCase("[N]")) {
							parseState = READ_N_X;
						} else {
							numberOfErrors++;
							System.err.println("INFO: [N] expected");
							parseState = READ_MARKER;
						}
						break;
					case READ_N_X:
						if (token.equalsIgnoreCase("[MARKER]")) {
							numberOfErrors++;
							System.err.println("INFO: data from n_x onwards missing");
							parseState = READ_X;
						} else {
							try {
								n_x = Float.parseFloat(token);
								parseState = READ_N_Y;
							} catch (NumberFormatException nfe) {
								numberOfErrors++;
								System.err.println("INFO: invalid (not a float) n_x coordinate");
								parseState = READ_MARKER;
							}
						}
						break;
					case READ_N_Y:
						if (token.equalsIgnoreCase("[MARKER]")) {
							numberOfErrors++;
							System.err.println("INFO: data from n_y onwards missing");
							parseState = READ_X;
						} else {
							try {
								n_y = Float.parseFloat(token);
								parseState = READ_N_Z;
							} catch (NumberFormatException nfe) {
								numberOfErrors++;
								System.err.println("INFO: invalid (not a float) n_y coordinate");
								parseState = READ_MARKER;
							}
						}
						break;
					case READ_N_Z:
						if (token.equalsIgnoreCase("[MARKER]")) {
							numberOfErrors++;
							System.err.println("INFO: datum n_z missing");
							parseState = READ_X;
						} else {
							try {
								n_z = Float.parseFloat(token);
								Vector p = new Vector(p_x, p_y, p_z);
								Vector u = new Vector(u_x, u_y, u_z);
								Vector v = new Vector(v_x, v_y, v_z);
								Vector normal = new Vector(n_x, n_y, n_z);
								Marker marker = new Marker(p, u, v, normal);
								buildInstance.add(marker);
								parseState = READ_MARKER;
							} catch (NumberFormatException nfe) {
								numberOfErrors++;
								System.err.println("INFO: invalid (not a float) n_z coordinate");
								parseState = READ_MARKER;
							}
						}
						break;
					}
					break;
				case EOF:
					switch (parseState) {
					case READ_MARKER:
						// orderly ending of file
						break;
					case READ_X:
						numberOfErrors++;
						System.err.println("INFO: data from p_x onwards missing");
						break;
					case READ_Y:
						numberOfErrors++;
						System.err.println("INFO: data from p_y onwards missing");
						break;
					case READ_Z:
						numberOfErrors++;
						System.err.println("INFO: data from p_z onwards missing");
						break;
					case READ_U:
						numberOfErrors++;
						System.err.println("INFO: data from [U} onwards missing");
						break;
					case READ_U_X:
						numberOfErrors++;
						System.err.println("INFO: data from u_x onwards missing");
						break;
					case READ_U_Y:
						numberOfErrors++;
						System.err.println("INFO: data from u_y onwards missing");
						break;
					case READ_U_Z:
						numberOfErrors++;
						System.err.println("INFO: data from u_z onwards missing");
						break;
					case READ_V:
						numberOfErrors++;
						System.err.println("INFO: data from [V} onwards missing");
						break;
					case READ_V_X:
						numberOfErrors++;
						System.err.println("INFO: data from v_x onwards missing");
						break;
					case READ_V_Y:
						numberOfErrors++;
						System.err.println("INFO: data from v_y onwards missing");
						break;
					case READ_V_Z:
						numberOfErrors++;
						System.err.println("INFO: data from v_z onwards missing");
						break;
					case READ_N:
						numberOfErrors++;
						System.err.println("INFO: data from [N} onwards missing");
						break;
					case READ_N_X:
						numberOfErrors++;
						System.err.println("INFO: data from n_x onwards missing");
						break;
					case READ_N_Y:
						numberOfErrors++;
						System.err.println("INFO: data from n_y onwards missing");
						break;
					case READ_N_Z:
						numberOfErrors++;
						System.err.println("INFO: datum n_z missing");
						break;
					}
					break;
				}
			} while (tokenType != EOF);
			s.close();
			System.err.println("INFO: number of parse errors is " +
			 numberOfErrors);
		} catch (FileNotFoundException e) {
			System.err.println("WARNING: unable to open file " +
			 animationFile.getName());
			return 0; // unable to open file so make sure that no parse error
					  // warning will be shown (this exception will be dealt
					  // with in a better way in a next version)
		}
		
		markers = buildInstance; // commit the build instance
		selectedMarker = null; // and nullify this variable
		Trail.generateTrail(); // regenerate the trail
		MAIN_SCREEN.findRepresentation(); // and find the new screen
										  // representation
		mainWindow.getTreeCanvas().repaint();
		return numberOfErrors;
	}
	
	// static method saveAnimationFile( is used to save the current animation
	// to a file
	public static void saveAnimationFile(File animationFile) {
		try {
			PrintWriter pw = new PrintWriter(animationFile);
			pw.println("! animation file for The Tree Project");
			pw.println("! see www.jogja.nl/manuals.html for details on the structure of this file");
			pw.println();
			pw.println("! [MARKER] p_x p_y p_z [U] u_x u_y u_z [V] v_x v_y v_z [N] n_x n_y n_z");
			pw.println();
			
			// iterate through the list of markers
			for (Marker marker : markers) {
				Vector u = marker.getU(), v = marker.getV(), normal = marker.getNormal();
				pw.print("[MARKER] " + marker.getX() + " " + marker.getY() + " " + marker.getZ());
				pw.print(" [U] " + u.getX() + " " + u.getY() + " " + u.getZ());
				pw.print(" [V] " + v.getX() + " " + v.getY() + " " + v.getZ());
				pw.print(" [N] " + normal.getX() + " " + normal.getY() + " " + normal.getZ());
				pw.println();
			}
			pw.close();
		} catch (FileNotFoundException e) {
			System.out.println("WARNING: unable to save animation");
		}
	}
	
	// this method is called when the user checks or unchecks "Show markers in
	// manual mode"
	public static void changeSmimmState(Checkbox smimm, MainWindow mainWindow) {
		showMarkersInManualMode = smimm.getState();
		mainWindow.getTreeCanvas().repaint();
	}
	
	// this method is called when the user checks or unchecks "Show markers in
	// animation mode"
	public static void changeSmiamState(Checkbox smiam) {
		showMarkersInAnimationMode = smiam.getState();
		// for the rest we do nothing because the change shall only visible
		// once the user starts an animation
	}
	
	// this method is called when the user clicks the "Set" button
	public static void setMarker(MainWindow mainWindow) {
		Marker m = new Marker();
		markers.add(m);
		Trail.generateTrail(); // regenerate the trail
		MAIN_SCREEN.findRepresentation(); // and find the new screen
										  // representation
		mainWindow.getTreeCanvas().repaint();
	}
	
	// this method is called when the user clicks the "Delete" button
	public static void deleteMarker(MainWindow mainWindow) {
		if (selectedMarker != null) {
			markers.remove(selectedMarker);
			selectedMarker = null;
			Trail.generateTrail(); // regenerate the trail
			MAIN_SCREEN.findRepresentation(); // and find the new screen
											  // representation
			mainWindow.getTreeCanvas().repaint();
		} else { // if no marker was selected, show an error message
			mainWindow.showDeleteErrorDialog();
		}
	}
	
	// this method is called when the user clicks the "Delete all" button
	public static void deleteAllMarkers(MainWindow mainWindow) {
		markers = new Markers(); // reset the markers object
		selectedMarker = null; // and nullify this variable
		Trail.generateTrail(); // regenerate the trail
		mainWindow.getTreeCanvas().repaint();
	}
	
}
